from typing import (TYPE_CHECKING, Any, Iterable, List, Optional, Sequence,
                    Tuple, Union)
from abc import abstractmethod

import numpy as np
from rasterio.features import rasterize
from shapely.ops import transform

from rastervision.core.box import Box
from rastervision.core.data.label import Labels
from rastervision.core.data.label.utils import discard_prediction_edges

if TYPE_CHECKING:
    from rastervision.core.data import (ClassConfig, CRSTransformer,
                                        VectorOutputConfig)


class SemanticSegmentationLabels(Labels):
    """Representation of Semantic Segmentation labels."""

    def __init__(self, extent: Box, num_classes: int, dtype: np.dtype):
        """Constructor.

        Args:
            extent (Box): The extent of the region to which
                the labels belong, in global coordinates.
            num_classes (int): Number of classes.
        """
        self.extent = extent
        self.num_classes = num_classes
        self.ymin, self.xmin, self.width, self.height = extent.to_xywh()
        self.dtype = dtype

    def _to_local_coords(self,
                         window: Union[Box, Tuple[int, int, int, int]]) -> Box:
        """Convert window to extent coordinates.

        Args:
            window (Union[Box, Tuple[int, int, int, int]]): A rastervision
                Box or a 4-tuple of (ymin, xmin, ymax, xmax).

        Returns:
            Box: Window in local coords.
        """
        ymin, xmin, ymax, xmax = window
        # convert to extent coords
        ymin_local, xmin_local = ymin - self.ymin, xmin - self.xmin
        ymax_local, xmax_local = ymax - self.ymin, xmax - self.xmin
        # clip negative values to zero
        ymin_local, xmin_local = max(0, ymin_local), max(0, xmin_local)
        ymax_local, xmax_local = max(0, ymax_local), max(0, xmax_local)
        # clip max values to array bounds
        ymax_local = min(self.height, ymax_local)
        xmax_local = min(self.width, xmax_local)
        return Box(ymin_local, xmin_local, ymax_local, xmax_local)

    @abstractmethod
    def __add__(self, other) -> 'SemanticSegmentationLabels':
        """Merge self with other labels."""
        pass

    def __setitem__(self, window: Box, values: np.ndarray) -> None:
        """Set labels for the given window."""
        self.add_window(window, values)

    @abstractmethod
    def __delitem__(self, window: Box) -> None:
        """Delete labels for the given window."""
        pass

    @abstractmethod
    def __getitem__(self, window: Box) -> np.ndarray:
        """Get labels for the given window."""
        pass

    @abstractmethod
    def add_window(self, window: Box, values: np.ndarray) -> List[Box]:
        """Set labels for the given window."""
        pass

    @abstractmethod
    def get_label_arr(self, window: Box,
                      null_class_id: int = -1) -> np.ndarray:
        """Get labels as array of class IDs.

        Note: The returned array is not guaranteed to be the same size as the
        input window.
        """
        pass

    def get_windows(self, **kwargs) -> List[Box]:
        """Generate sliding windows over the local extent. The keyword args
        are passed to Box.get_windows() and can therefore be used to control
        the specifications of the windows.

        If the keyword args do not contain size, a list of length 1,
        containing the full extent is returned.
        """
        size: Optional[int] = kwargs.pop('size', None)
        if size is None:
            return [self.extent]
        return self.extent.get_windows(size, size, **kwargs)

    def filter_by_aoi(self, aoi_polygons: list, null_class_id: int,
                      **kwargs) -> 'SemanticSegmentationLabels':
        """Keep only the values that lie inside the AOI."""
        if not aoi_polygons:
            return self
        for window in self.get_windows(**kwargs):
            self._filter_window_by_aoi(window, aoi_polygons, null_class_id)
        return self

    @abstractmethod
    def mask_fill(self, window: Box, mask: np.ndarray,
                  fill_value: Any) -> None:
        """Given a window and a binary mask, set all the pixels in the window
        for which the mask is ON to the fill_value.
        """
        pass

    def _filter_window_by_aoi(self, window: Box, aoi_polygons: list,
                              null_class_id: int) -> None:
        window_geom = window.to_shapely()
        label_arr = self[window]

        # For each aoi_polygon, intersect with window, and
        # put in window frame of reference.
        window_aois = []
        for aoi in aoi_polygons:
            window_aoi = aoi.intersection(window_geom)
            if not window_aoi.is_empty:

                def transform_shape(x, y, z=None):
                    return (x - window.xmin, y - window.ymin)

                window_aoi = transform(transform_shape, window_aoi)
                window_aois.append(window_aoi)

        # If window does't overlap with any AOI, then it won't be in
        # new_labels.
        if window_aois:
            # If window intersects with AOI, set pixels outside the
            # AOI polygon to 0 so they are ignored during eval.
            mask = rasterize(
                [(p, 0) for p in window_aois],
                out_shape=label_arr.shape[-2:],
                fill=1,
                dtype=np.uint8)
            mask = mask.astype(bool)
            self.mask_fill(window, mask, null_class_id)
        else:
            del self[window]

    @classmethod
    def make_empty(cls, extent: Box, num_classes: int, smooth: bool = False
                   ) -> Union['SemanticSegmentationDiscreteLabels',
                              'SemanticSegmentationSmoothLabels']:
        """Instantiate an empty instance.

        Args:
            extent (Box): The extent of the region to which the labels belong,
                in global coordinates.
            num_classes (int): Number of classes.
            smooth (bool, optional): If True, creates a
                SemanticSegmentationSmoothLabels object. If False, creates a
                SemanticSegmentationDiscreteLabels object. Defaults to False.

        Returns:
            Union[SemanticSegmentationDiscreteLabels,
            SemanticSegmentationSmoothLabels]: If smooth=True, returns a
                SemanticSegmentationSmoothLabels. Otherwise, a
                SemanticSegmentationDiscreteLabels.

        Raises:
            ValueError: if num_classes and extent are not specified, but
                smooth=True.
        """
        if not smooth:
            return SemanticSegmentationDiscreteLabels.make_empty(
                extent=extent, num_classes=num_classes)
        else:
            return SemanticSegmentationSmoothLabels.make_empty(
                extent=extent, num_classes=num_classes)

    @classmethod
    def from_predictions(cls,
                         windows: Iterable['Box'],
                         predictions: Iterable[Any],
                         extent: Box,
                         num_classes: int,
                         smooth: bool = False,
                         crop_sz: Optional[int] = None
                         ) -> Union['SemanticSegmentationDiscreteLabels',
                                    'SemanticSegmentationSmoothLabels']:
        """Instantiate from windows and their corresponding predictions.

        Args:
            windows (Iterable[Box]): Boxes in pixel coords, specifying chips
                in the raster.
            predictions (Iterable[Any]): The model predictions for each chip
                specified by the windows.
            extent (Box): The extent of the region to which the labels belong,
                in global coordinates.
            num_classes (int): Number of classes.
            crop_sz (Optional[int]): Number of rows/columns of pixels from the
                edge of prediction windows to discard. This is useful because
                predictions near edges tend to be lower quality and can result
                in very visible artifacts near the edges of chips. This should
                only be used if the given windows represent a sliding-window
                grid over the scene extent with overlap between adjacent
                windows. Defaults to None.

        Returns:
            Union[SemanticSegmentationDiscreteLabels,
            SemanticSegmentationSmoothLabels]: If smooth=True, returns a
                SemanticSegmentationSmoothLabels. Otherwise, a
                SemanticSegmentationDiscreteLabels.
        """
        labels = cls.make_empty(extent, num_classes, smooth=smooth)
        labels.add_predictions(windows, predictions, crop_sz=crop_sz)
        return labels

    def add_predictions(self,
                        windows: Iterable['Box'],
                        predictions: Iterable[Any],
                        crop_sz: Optional[int] = None) -> None:
        """Populate predictions.

        Args:
            windows (Iterable[Box]): Boxes in pixel coords, specifying chips
                in the raster.
            predictions (Iterable[Any]): The model predictions for each chip
                specified by the windows.
            crop_sz (Optional[int]): Number of rows/columns of pixels from the
                edge of prediction windows to discard. This is useful because
                predictions near edges tend to be lower quality and can result
                in very visible artifacts near the edges of chips. This should
                only be used if the given windows represent a sliding-window
                grid over the scene extent with overlap between adjacent
                windows. Defaults to None.
        """
        if crop_sz is not None:
            windows, predictions = discard_prediction_edges(
                windows, predictions, crop_sz)
        # If predictions is tqdm-wrapped, it needs to be the first arg to zip()
        # or the progress bar won't terminate with the correct count.
        for prediction, window in zip(predictions, windows):
            self[window] = prediction


class SemanticSegmentationDiscreteLabels(SemanticSegmentationLabels):
    """Vote-counts for each pixel belonging to each class.

    Maintains a num_classes x H x W array where value_{ijk} represents how
    many times pixel_{jk} has been classified as class i. A label array can be
    obtained from this by argmax'ing along the class dimension. Can also be
    turned into a score converting counts to probabilities.
    """

    def __init__(self, extent: Box, num_classes: int, dtype: Any = np.uint8):
        """Constructor.

        Args:
            extent (Box): The extent of the region to which
                the labels belong, in global coordinates.
            num_classes (int): Number of classes.
            dtype (Any): dtype of the counts array. Defaults to np.uint8.
        """
        super().__init__(extent, num_classes, dtype)

        self.pixel_counts = np.zeros(
            (self.num_classes, self.height, self.width), dtype=self.dtype)
        # track which pixels have been hit at all
        self.hit_mask = np.zeros((self.height, self.width), dtype=bool)

    def __add__(self, other: 'SemanticSegmentationDiscreteLabels'
                ) -> 'SemanticSegmentationDiscreteLabels':
        """Merge self with other labels by adding the pixel counts."""
        if self.extent != other.extent:
            raise ValueError('Cannot add labels with unqeual extents.')

        self.pixel_counts += other.pixel_counts
        return self

    def __eq__(self, other: 'SemanticSegmentationDiscreteLabels') -> bool:
        if not isinstance(other, SemanticSegmentationDiscreteLabels):
            return False
        if self.extent != other.extent:
            return False
        mask_equal = np.all(self.hit_mask == other.hit_mask)
        if not mask_equal:
            return False
        counts_equal = np.all(self.pixel_counts == other.pixel_counts)
        return counts_equal

    def __delitem__(self, window: Box) -> None:
        """Reset counts to zero for pixels in the window."""
        y0, x0, y1, x1 = self._to_local_coords(window)
        self.pixel_counts[..., y0:y1, x0:x1] = 0
        self.hit_mask[y0:y1, x0:x1] = False

    def __getitem__(self, window: Box) -> np.ndarray:
        return self.get_label_arr(window)

    def add_window(self, window: Box, pixel_class_ids: np.ndarray) -> None:
        window_local = self._to_local_coords(window)
        dst_yslice, dst_xslice = window_local.to_slices()
        src_yslice, src_xslice = window_local.to_offsets(window).to_slices()

        pixel_class_ids = pixel_class_ids.astype(self.dtype)
        pixel_class_ids = pixel_class_ids[..., src_yslice, src_xslice]
        window_pixel_counts = self.pixel_counts[:, dst_yslice, dst_xslice]
        for ch_class_id, ch in enumerate(window_pixel_counts):
            ch[pixel_class_ids == ch_class_id] += 1
        self.hit_mask[dst_yslice, dst_xslice] = True

    def get_label_arr(self, window: Box,
                      null_class_id: int = -1) -> np.ndarray:
        """Get labels as array of class IDs.

        Returns null_class_id for pixels for which there is no data.
        """
        y0, x0, y1, x1 = self._to_local_coords(window)
        label_arr = self.pixel_counts[..., y0:y1, x0:x1].argmax(axis=0)
        hit_mask = self.hit_mask[y0:y1, x0:x1]
        return np.where(hit_mask, label_arr, null_class_id)

    def get_score_arr(self, window: Box) -> np.ndarray:
        """Get array of pixel scores."""
        y0, x0, y1, x1 = self._to_local_coords(window)
        class_counts = self.pixel_counts[..., y0:y1, x0:x1]
        scores = class_counts / class_counts.sum(axis=0)
        return scores

    def mask_fill(self, window: Box, mask: np.ndarray,
                  fill_value: Any) -> None:
        """Set fill_value'th class ID's count to 1 and all others to zero."""
        class_id = fill_value
        y0, x0, y1, x1 = self._to_local_coords(window)
        h, w = y1 - y0, x1 - x0
        mask = mask[:h, :w]
        self.pixel_counts[:, y0:y1, x0:x1][..., mask] = 0
        self.pixel_counts[class_id, y0:y1, x0:x1][mask] = 1

    @classmethod
    def make_empty(cls, extent: Box,
                   num_classes: int) -> 'SemanticSegmentationDiscreteLabels':
        """Instantiate an empty instance."""
        return cls(extent=extent, num_classes=num_classes)

    @classmethod
    def from_predictions(cls,
                         windows: Iterable['Box'],
                         predictions: Iterable[Any],
                         extent: Box,
                         num_classes: int,
                         crop_sz: Optional[int] = None
                         ) -> Union['SemanticSegmentationDiscreteLabels',
                                    'SemanticSegmentationSmoothLabels']:
        labels = cls.make_empty(extent, num_classes)
        labels.add_predictions(windows, predictions, crop_sz=crop_sz)
        return labels

    def save(self,
             uri: str,
             crs_transformer: 'CRSTransformer',
             class_config: 'ClassConfig',
             tmp_dir: Optional[str] = None,
             save_as_rgb: bool = False,
             raster_output: bool = True,
             rasterio_block_size: int = 512,
             vector_outputs: Optional[Sequence['VectorOutputConfig']] = None,
             profile_overrides: Optional[dict] = None) -> None:
        """Save labels as a raster and/or vectors.

        If URI is remote, all files will be first written locally and then
        uploaded to the URI.

        Args:
            uri (str): URI of directory in which to save all output files.
            crs_transformer (CRSTransformer): CRSTransformer to configure CRS
                and affine transform of the output GeoTiff.
            class_config (ClassConfig): The ClassConfig.
            tmp_dir (Optional[str], optional): Temporary directory to use. If
                None, will be auto-generated. Defaults to None.
            save_as_rgb (bool, optional): If True, Saves labels as an RGB
                image, using the class-color mapping in the class_config.
                Defaults to False.
            raster_output (bool, optional): If True, saves labels as a raster
                of class IDs (one band). Defaults to True.
            rasterio_block_size (int, optional): Value to set blockxsize and
                blockysize to. Defaults to 512.
            vector_outputs (Optional[Sequence[VectorOutputConfig]], optional):
                List of VectorOutputConfig's containing vectorization
                configuration information. Only classes for which a
                VectorOutputConfig is specified will be saved as vectors.
                If None, no vector outputs will be produced. Defaults to None.
            profile_overrides (Optional[dict], optional): This can be used to
                arbitrarily override properties in the profile used to create
                the output GeoTiff. Defaults to None.
        """
        from rastervision.core.data import SemanticSegmentationLabelStore

        label_store = SemanticSegmentationLabelStore(
            uri=uri,
            extent=self.extent,
            crs_transformer=crs_transformer,
            class_config=class_config,
            tmp_dir=tmp_dir,
            save_as_rgb=save_as_rgb,
            discrete_output=raster_output,
            smooth_output=False,
            rasterio_block_size=rasterio_block_size,
            vector_outputs=vector_outputs)
        label_store.save(self, profile=profile_overrides)


class SemanticSegmentationSmoothLabels(SemanticSegmentationLabels):
    """Membership-scores for each pixel for each class.

    Maintains a num_classes x H x W array where value_{ijk} represents the
    probability (or some other measure) of pixel_{jk} belonging to class i.
    A discrete label array can be obtained from this by argmax'ing along the
    class dimension.
    """

    def __init__(self,
                 extent: Box,
                 num_classes: int,
                 dtype: Any = np.float16,
                 dtype_hits: Any = np.uint8):
        """Constructor.

        Args:
            extent (Box): The extent of the region to which
                the labels belong, in global coordinates.
            num_classes (int): Number of classes.
            dtype (Any): dtype of the scores array. Defaults to np.float16.
            dtype_hits (Any): dtype of the hits array. Defaults to np.uint8.
        """
        super().__init__(extent, num_classes, dtype)

        self.pixel_scores = np.zeros(
            (self.num_classes, self.height, self.width), dtype=self.dtype)
        self.pixel_hits = np.zeros((self.height, self.width), dtype=dtype_hits)

    def __add__(self, other: 'SemanticSegmentationSmoothLabels'
                ) -> 'SemanticSegmentationSmoothLabels':
        """Merge self with other by adding pixel scores and hits."""
        if self.extent != other.extent:
            raise ValueError('Cannot add labels with unqeual extents.')

        self.pixel_scores += other.pixel_scores
        self.pixel_hits += other.pixel_hits
        return self

    def __eq__(self, other: 'SemanticSegmentationSmoothLabels') -> bool:
        if not isinstance(other, SemanticSegmentationSmoothLabels):
            return False
        if self.extent != other.extent:
            return False
        scores_equal = np.allclose(self.pixel_scores, other.pixel_scores)
        hits_equal = np.array_equal(self.pixel_hits, other.pixel_hits)
        return (scores_equal and hits_equal)

    def __delitem__(self, window: Box) -> None:
        """Reset scores and hits to zero for pixels in the window."""
        y0, x0, y1, x1 = self._to_local_coords(window)
        self.pixel_scores[..., y0:y1, x0:x1] = 0
        self.pixel_hits[..., y0:y1, x0:x1] = 0

    def __getitem__(self, window: Box) -> np.ndarray:
        return self.get_score_arr(window)

    def add_window(self, window: Box, pixel_class_scores: np.ndarray) -> None:
        window_local = self._to_local_coords(window)
        dst_yslice, dst_xslice = window_local.to_slices()
        src_yslice, src_xslice = window_local.to_offsets(window).to_slices()

        pixel_class_scores = pixel_class_scores.astype(self.dtype)
        pixel_class_scores = pixel_class_scores[..., src_yslice, src_xslice]
        self.pixel_scores[..., dst_yslice, dst_xslice] += pixel_class_scores
        self.pixel_hits[dst_yslice, dst_xslice] += 1

    def get_score_arr(self, window: Box) -> np.ndarray:
        """Get array of pixel scores."""
        y0, x0, y1, x1 = self._to_local_coords(window)
        scores = self.pixel_scores[..., y0:y1, x0:x1]
        hits = self.pixel_hits[y0:y1, x0:x1]
        avg_scores = scores / hits
        return avg_scores

    def get_label_arr(self, window: Box,
                      null_class_id: int = -1) -> np.ndarray:
        """Get labels as array of class IDs.

        Returns null_class_id for pixels for which there is no data.
        """
        avg_scores = self.get_score_arr(window)
        label_arr = np.argmax(avg_scores, axis=0)
        mask = np.isnan(avg_scores[0])
        return np.where(mask, null_class_id, label_arr)

    def mask_fill(self, window: Box, mask: np.ndarray,
                  fill_value: Any) -> None:
        """Set fill_value'th class ID's score to 1 and all others to zero."""
        class_id = fill_value
        y0, x0, y1, x1 = self._to_local_coords(window)
        h, w = y1 - y0, x1 - x0
        mask = mask[:h, :w]
        self.pixel_scores[..., y0:y1, x0:x1][..., mask] = 0
        self.pixel_scores[class_id, y0:y1, x0:x1][mask] = 1
        self.pixel_hits[y0:y1, x0:x1][mask] = 1

    @classmethod
    def make_empty(cls, extent: Box,
                   num_classes: int) -> 'SemanticSegmentationSmoothLabels':
        """Instantiate an empty instance."""
        return cls(extent=extent, num_classes=num_classes)

    @classmethod
    def from_predictions(cls,
                         windows: Iterable['Box'],
                         predictions: Iterable[Any],
                         extent: Box,
                         num_classes: int,
                         crop_sz: Optional[int] = None
                         ) -> Union['SemanticSegmentationDiscreteLabels',
                                    'SemanticSegmentationSmoothLabels']:
        labels = cls.make_empty(extent, num_classes)
        labels.add_predictions(windows, predictions, crop_sz=crop_sz)
        return labels

    def save(self,
             uri: str,
             crs_transformer: 'CRSTransformer',
             class_config: 'ClassConfig',
             tmp_dir: Optional[str] = None,
             save_as_rgb: bool = False,
             discrete_output: bool = True,
             smooth_output: bool = True,
             smooth_as_uint8: bool = False,
             rasterio_block_size: int = 512,
             vector_outputs: Optional[Sequence['VectorOutputConfig']] = None,
             profile_overrides: Optional[dict] = None) -> None:
        """Save labels as rasters and/or vectors.

        If URI is remote, all files will be first written locally and then
        uploaded to the URI.

        Args:
            uri (str): URI of directory in which to save all output files.
            crs_transformer (CRSTransformer): CRSTransformer to configure CRS
                and affine transform of the output GeoTiff(s).
            class_config (ClassConfig): The ClassConfig.
            tmp_dir (Optional[str], optional): Temporary directory to use. If
                None, will be auto-generated. Defaults to None.
            save_as_rgb (bool, optional): If True, saves labels as an RGB
                image, using the class-color mapping in the class_config.
                Defaults to False.
            discrete_output (bool, optional): If True, saves labels as a raster
                of class IDs (one band). Defaults to True.
            smooth_output (bool, optional): If True, saves labels as a raster
                of class scores (one band for each class). Defaults to True.
            smooth_as_uint8 (bool, optional): If True, stores smooth class
                scores as np.uint8 (0-255) values rather than as np.float32
                discrete labels, to help save memory/disk space.
                Defaults to False.
            rasterio_block_size (int, optional): Value to set blockxsize and
                blockysize to. Defaults to 512.
            vector_outputs (Optional[Sequence[VectorOutputConfig]], optional):
                List of VectorOutputConfig's containing vectorization
                configuration information. Only classes for which a
                VectorOutputConfig is specified will be saved as vectors.
                If None, no vector outputs will be produced. Defaults to None.
            profile_overrides (Optional[dict], optional): This can be used to
                arbitrarily override properties in the profile used to create
                the output GeoTiff(s). Defaults to None.
        """
        from rastervision.core.data import SemanticSegmentationLabelStore

        label_store = SemanticSegmentationLabelStore(
            uri=uri,
            extent=self.extent,
            crs_transformer=crs_transformer,
            class_config=class_config,
            tmp_dir=tmp_dir,
            save_as_rgb=save_as_rgb,
            discrete_output=discrete_output,
            smooth_output=smooth_output,
            smooth_as_uint8=smooth_as_uint8,
            rasterio_block_size=rasterio_block_size,
            vector_outputs=vector_outputs)
        label_store.save(self, profile=profile_overrides)



# flake8: noqa

from rastervision.core.evaluation.evaluation_item import *
from rastervision.core.evaluation.class_evaluation_item import *
from rastervision.core.evaluation.evaluator import *
from rastervision.core.evaluation.evaluator_config import *
from rastervision.core.evaluation.classification_evaluation import *
from rastervision.core.evaluation.classification_evaluator import *
from rastervision.core.evaluation.classification_evaluator_config import *
from rastervision.core.evaluation.chip_classification_evaluation import *
from rastervision.core.evaluation.chip_classification_evaluator import *
from rastervision.core.evaluation.chip_classification_evaluator_config import *
from rastervision.core.evaluation.semantic_segmentation_evaluation import *
from rastervision.core.evaluation.semantic_segmentation_evaluator import *
from rastervision.core.evaluation.semantic_segmentation_evaluator_config import *
from rastervision.core.evaluation.object_detection_evaluation import *
from rastervision.core.evaluation.object_detection_evaluator import *
from rastervision.core.evaluation.object_detection_evaluator_config import *

__all__ = [
    EvaluationItem.__name__,
    ClassEvaluationItem.__name__,
    Evaluator.__name__,
    EvaluatorConfig.__name__,
    ClassificationEvaluation.__name__,
    ClassificationEvaluator.__name__,
    ClassificationEvaluatorConfig.__name__,
    ChipClassificationEvaluation.__name__,
    ChipClassificationEvaluator.__name__,
    ChipClassificationEvaluatorConfig.__name__,
    SemanticSegmentationEvaluation.__name__,
    SemanticSegmentationEvaluator.__name__,
    SemanticSegmentationEvaluatorConfig.__name__,
    ObjectDetectionEvaluation.__name__,
    ObjectDetectionEvaluator.__name__,
    ObjectDetectionEvaluatorConfig.__name__,
]



from typing import TYPE_CHECKING, Dict, List, Optional
from rastervision.core.data.raster_transformer import RasterTransformer

if TYPE_CHECKING:
    import numpy as np


class ReclassTransformer(RasterTransformer):
    """Maps class IDs in a label raster to other values."""

    def __init__(self, mapping: Dict[int, int]):
        """Construct a new ReclassTransformer.

        Args:
            mapping: (dict) Remapping dictionary
        """
        self.mapping = mapping

    def transform(self,
                  chip: 'np.ndarray',
                  channel_order: Optional[List[int]] = None):
        """Transform a chip.

        Reclassify a label raster using the given mapping.

        Args:
            chip: ndarray of shape [height, width, channels] This is assumed to already
                have the channel_order applied to it if channel_order is set. In other
                words, channels should be equal to len(channel_order).
            channel_order: list of indices of channels that were extracted from the
                raw imagery.

        Returns:
            [height, width, channels] numpy array

        """
        masks = []
        for (value_from, value_to) in self.mapping.items():
            mask = (chip == value_from)
            masks.append((mask, value_to))
        for (mask, value_to) in masks:
            chip[mask] = value_to

        return chip



from typing import Any, Optional, Tuple

import numpy as np

from rastervision.core.box import Box
from rastervision.core.data.label import ObjectDetectionLabels
from rastervision.core.data.label_source import LabelSource
from rastervision.core.data.vector_source import VectorSource


class ObjectDetectionLabelSource(LabelSource):
    """A read-only label source for object detection."""

    def __init__(self,
                 vector_source: VectorSource,
                 extent: Box,
                 ioa_thresh: Optional[float] = None,
                 clip: bool = False):
        """Constructor.

        Args:
            vector_source (VectorSource): A VectorSource.
            extent (Box): Box used to filter the labels by extent.
            ioa_thresh (Optional[float], optional): IOA threshold to apply when
                retieving labels for a window. Defaults to None.
            clip (bool, optional): Clip bounding boxes to window limits when
                retrieving labels for a window. Defaults to False.
        """
        geojson = vector_source.get_geojson()
        self.validate_geojson(geojson)
        self.labels = ObjectDetectionLabels.from_geojson(
            geojson, extent=extent)
        self._extent = extent
        self.ioa_thresh = ioa_thresh if ioa_thresh is not None else 1e-6
        self.clip = clip

    def get_labels(self,
                   window: Box = None,
                   ioa_thresh: float = 1e-6,
                   clip: bool = False) -> ObjectDetectionLabels:
        """Get labels (in global coords) for a window.

        Args:
            window (Box): Window coords.

        Returns:
            ObjectDetectionLabels: Labels with sufficient overlap with the
                window. The returned labels are in global coods
                (i.e. coords wihtin the full extent).
        """
        if window is None:
            return self.labels
        window = window.shift_origin(self.extent)
        return ObjectDetectionLabels.get_overlapping(
            self.labels, window, ioa_thresh=ioa_thresh, clip=clip)

    def __getitem__(self, key: Any) -> Tuple[np.ndarray, np.ndarray, str]:
        """Get labels (in window coords) for a window.

        Returns a 3-tuple: (npboxes, class_ids, box_format).
        - npboxes is a float np.ndarray of shape (num_boxes, 4) representing
        pixel coords of bounding boxes in the form [ymin, xmin, ymax, xmax].
        - class_ids is a np.ndarray of shape (num_boxes,) representing the
        class labels for each of the boxes.
        - box_format is the format of npboxes which, in this case, is always
        'yxyx'.

        Args:
            window (Box): Window coords.

        Returns:
            Tuple[np.ndarray, np.ndarray, str]: 3-tuple of
                (npboxes, class_ids, box_format). The returned npboxes are in
                window coords (i.e. coords within the window).
        """
        if isinstance(key, Box):
            window = key
            labels = self.get_labels(
                window, ioa_thresh=self.ioa_thresh, clip=self.clip)
            class_ids = labels.get_class_ids()
            npboxes = labels.get_npboxes()
            npboxes = ObjectDetectionLabels.global_to_local(npboxes, window)
            return npboxes, class_ids, 'yxyx'
        elif isinstance(key, slice):
            key = [key]
        elif isinstance(key, tuple):
            pass
        else:
            raise TypeError('Unsupported key type.')
        slices = list(key)
        assert 1 <= len(slices) <= 2
        assert all(s is not None for s in slices)
        assert isinstance(slices[0], slice)
        if len(slices) == 1:
            h, = slices
            w = slice(None, None)
        else:
            assert isinstance(slices[1], slice)
            h, w = slices

        if any(x is not None and x < 0
               for x in [h.start, h.stop, w.start, w.stop]):
            raise NotImplementedError()

        ymin, xmin, ymax, xmax = self.extent
        _ymin = 0 if h.start is None else h.start
        _xmin = 0 if w.start is None else w.start
        _ymax = ymax if h.stop is None else h.stop
        _xmax = xmax if w.stop is None else w.stop
        window = Box(_ymin, _xmin, _ymax, _xmax)

        npboxes, class_ids, fmt = self[window]

        # rescale if steps specified
        if h.step is not None:
            # assume fmt='yxyx'
            npboxes[:, [0, 2]] /= h.step
        if w.step is not None:
            # assume fmt='yxyx'
            npboxes[:, [1, 3]] /= w.step

        return npboxes, class_ids, fmt

    def validate_geojson(self, geojson: dict) -> None:
        for f in geojson['features']:
            geom_type = f.get('geometry', {}).get('type', '')
            if 'Point' in geom_type or 'LineString' in geom_type:
                raise ValueError(
                    'LineStrings and Points are not supported '
                    'in ChipClassificationLabelSource. Use BufferTransformer '
                    'to buffer them into Polygons.')
        for f in geojson['features']:
            if f.get('properties', {}).get('class_id') is None:
                raise ValueError('All GeoJSON features must have a class_id '
                                 'field in their properties.')

    @property
    def extent(self) -> Box:
        return self._extent



from typing import TYPE_CHECKING
from rastervision.core.data.vector_source import (VectorSourceConfig,
                                                  GeoJSONVectorSource)
from rastervision.pipeline.config import register_config, Field

if TYPE_CHECKING:
    from rastervision.core.data import (ClassConfig, CRSTransformer)


@register_config('geojson_vector_source')
class GeoJSONVectorSourceConfig(VectorSourceConfig):
    """Configure a :class:`.GeoJSONVectorSource`."""

    uri: str = Field(..., description='The URI of a GeoJSON file.')
    ignore_crs_field: bool = False

    def build(self,
              class_config: 'ClassConfig',
              crs_transformer: 'CRSTransformer',
              use_transformers: bool = True) -> GeoJSONVectorSource:
        if use_transformers:
            transformers = [
                tf.build(class_config=class_config) for tf in self.transformers
            ]
        else:
            transformers = []

        return GeoJSONVectorSource(
            uri=self.uri,
            ignore_crs_field=self.ignore_crs_field,
            crs_transformer=crs_transformer,
            vector_transformers=transformers)



# flake8: noqa

from rastervision.core.data.label.labels import *
from rastervision.core.data.label.chip_classification_labels import *
from rastervision.core.data.label.semantic_segmentation_labels import *
from rastervision.core.data.label.object_detection_labels import *
from rastervision.core.data.label.utils import *

__all__ = [
    Labels.__name__,
    SemanticSegmentationLabels.__name__,
    SemanticSegmentationDiscreteLabels.__name__,
    SemanticSegmentationSmoothLabels.__name__,
    ObjectDetectionLabels.__name__,
    ChipClassificationLabels.__name__,
    ClassificationLabel.__name__,
]



from typing import (TYPE_CHECKING, Any, Dict, Iterable, List, Optional,
                    Sequence, Tuple)
from dataclasses import dataclass

import numpy as np

from rastervision.core.box import Box
from rastervision.core.data.label import Labels

if TYPE_CHECKING:
    from rastervision.core.data import (ClassConfig, CRSTransformer)
    from shapely.geometry import Polygon


@dataclass
class ClassificationLabel:
    class_id: int
    scores: Optional[Sequence[float]] = None

    def __iter__(self):
        return iter((self.class_id, self.scores))


class ChipClassificationLabels(Labels):
    """Represents a spatial grid of cells associated with classes."""

    def __init__(self,
                 cell_to_label: Optional[Dict[Box, Tuple[int, Optional[
                     Sequence[float]]]]] = None):
        if cell_to_label is None:
            cell_to_label = {}

        self.cell_to_label = {
            c: ClassificationLabel(*v)
            for c, v in cell_to_label.items()
        }

    def __len__(self) -> int:
        return len(self.cell_to_label)

    def __eq__(self, other: 'ChipClassificationLabels') -> bool:
        return (isinstance(other, ChipClassificationLabels)
                and self.cell_to_label == other.cell_to_label)

    def __add__(self, other: 'ChipClassificationLabels'
                ) -> 'ChipClassificationLabels':
        result = ChipClassificationLabels()
        result.extend(self)
        result.extend(other)
        return result

    def __contains__(self, cell: Box) -> bool:
        return cell in self.cell_to_label

    def __getitem__(self, cell: Box) -> ClassificationLabel:
        return self.cell_to_label[cell]

    def __setitem__(self, window: Box,
                    value: Tuple[int, Optional[Sequence[float]]]):
        class_id, scores = value
        self.set_cell(window, class_id, scores=scores)

    @classmethod
    def from_predictions(cls, windows: Iterable['Box'],
                         predictions: Iterable[Any]) -> 'Labels':
        """Overrid to convert predictions to (class_id, scores) pairs."""
        predictions = ((np.argmax(p), p) for p in predictions)
        return super().from_predictions(windows, predictions)

    @classmethod
    def make_empty(cls) -> 'ChipClassificationLabels':
        return ChipClassificationLabels()

    def filter_by_aoi(self, aoi_polygons: Iterable['Polygon']):
        result = ChipClassificationLabels()
        for cell in self.cell_to_label:
            cell_box = Box(*cell)
            cell_poly = cell_box.to_shapely()
            for aoi in aoi_polygons:
                if cell_poly.within(aoi):
                    (class_id, scores) = self.cell_to_label[cell]
                    result.set_cell(cell_box, class_id, scores)
        return result

    def set_cell(self,
                 cell: Box,
                 class_id: int,
                 scores: Optional['np.ndarray'] = None) -> None:
        """Set cell and its class_id.

        Args:
            cell: (Box)
            class_id: int
            scores: 1d numpy array of probabilities for each class
        """
        if scores is not None:
            scores = list(map(lambda x: float(x), list(scores)))
        class_id = int(class_id)
        self.cell_to_label[cell] = ClassificationLabel(class_id, scores)

    def get_cell_class_id(self, cell: Box) -> int:
        """Return class_id for a cell.

        Args:
            cell: (Box)
        """
        result = self.cell_to_label.get(cell)
        if result is not None:
            return result.class_id
        else:
            return None

    def get_cell_scores(self, cell: Box) -> Optional[Sequence[float]]:
        """Return scores for a cell.

        Args:
            cell: (Box)
        """
        result = self.cell_to_label.get(cell)
        if result is not None:
            return result.score
        else:
            return None

    def get_singleton_labels(self, cell: Box):
        """Return Labels object representing a single cell.

        Args:
            cell: (Box)
        """
        return ChipClassificationLabels({cell: self[cell]})

    def get_cells(self) -> List[Box]:
        """Return list of all cells (list of Box)."""
        return list(self.cell_to_label.keys())

    def get_class_ids(self) -> List[int]:
        """Return list of class_ids for all cells."""
        return [label.class_id for label in self.cell_to_label.values()]

    def get_scores(self) -> List[Optional[Sequence[float]]]:
        """Return list of scores for all cells."""
        return [label.scores for label in self.cell_to_label.values()]

    def get_values(self) -> List[ClassificationLabel]:
        """Return list of class_ids and scores for all cells."""
        return list(self.cell_to_label.values())

    def extend(self, labels: 'ChipClassificationLabels') -> None:
        """Adds cells contained in labels.

        Args:
            labels: ChipClassificationLabels
        """
        for cell in labels.get_cells():
            self.set_cell(cell, *labels[cell])

    def save(self, uri: str, class_config: 'ClassConfig',
             crs_transformer: 'CRSTransformer') -> None:
        """Save labels as a GeoJSON file.

        Args:
            uri (str): URI of the output file.
            class_config (ClassConfig): ClassConfig to map class IDs to names.
            crs_transformer (CRSTransformer): CRSTransformer to convert from
                pixel-coords to map-coords before saving.
        """
        from rastervision.core.data import ChipClassificationGeoJSONStore

        label_store = ChipClassificationGeoJSONStore(
            uri=uri,
            class_config=class_config,
            crs_transformer=crs_transformer)
        label_store.save(self)



from typing import TYPE_CHECKING, List, Iterable, Optional, Tuple

from rastervision.pipeline.config import register_config, Config

if TYPE_CHECKING:
    from rastervision.core.analyzer.analyzer import Analyzer  # noqa


@register_config('analyzer')
class AnalyzerConfig(Config):
    """Configure an :class:`.Analyzer`."""

    def build(self, scene_group: Optional[Tuple[str, Iterable[str]]] = None
              ) -> 'Analyzer':
        pass

    def get_bundle_filenames(self) -> List[str]:
        """Returns the names of files that should be included in a model bundle.

        The files are assumed to be in the analyze/ directory generated by the analyze
        command. Note that only the names, not the full paths should be returned.
        """
        return []



from typing import (Any, List, Optional)

import numpy as np

from rastervision.core.box import Box
from rastervision.core.data import ClassConfig
from rastervision.core.data.label import SemanticSegmentationLabels
from rastervision.core.data.label_source.label_source import LabelSource
from rastervision.core.data.raster_source import RasterSource


def fill_edge(label_arr: np.ndarray, window: Box, extent: Box,
              fill_value: int) -> np.ndarray:
    """If window goes over the edge of the extent, buffer with fill_value."""
    if window.ymax <= extent.ymax and window.xmax <= extent.xmax:
        return label_arr

    x = np.full(window.size, fill_value)
    ylim = extent.ymax - window.ymin
    xlim = extent.xmax - window.xmin
    x[0:ylim, 0:xlim] = label_arr[0:ylim, 0:xlim]
    return x


class SemanticSegmentationLabelSource(LabelSource):
    """A read-only label source for semantic segmentation."""

    def __init__(self, raster_source: RasterSource, class_config: ClassConfig):
        """Constructor.

        Args:
            raster_source (RasterSource): A raster source that returns a single
                channel raster with class_ids as values.
            null_class_id (int): the null class id used as fill values for when
                windows go over the edge of the label array. This can be
                retrieved using class_config.null_class_id.
        """
        self.raster_source = raster_source
        self.class_config = class_config

    def enough_target_pixels(self, window: Box, target_count_threshold: int,
                             target_classes: List[int]) -> bool:
        """Check if window contains enough pixels of the given target classes.

        Args:
             window: The larger window from-which the sub-window will
                  be clipped.
             target_count_threshold:  Minimum number of target pixels.
             target_classes: The classes of interest.  The given
                  window is examined to make sure that it contains a
                  sufficient number of target pixels.
        Returns:
             True (the window does contain interesting pixels) or False.
        """
        label_arr = self.get_label_arr(window)

        target_count = 0
        for class_id in target_classes:
            target_count += (label_arr == class_id).sum()

        return target_count >= target_count_threshold

    def get_labels(self,
                   window: Optional[Box] = None) -> SemanticSegmentationLabels:
        """Get labels for a window.

        Args:
            window (Optional[Box], optional): Window to get labels for. If
            None, returns labels covering the full extent of the scene.

        Returns:
            SemanticSegmentationLabels: The labels.
        """
        if window is None:
            window = self.extent

        label_arr = self.get_label_arr(window)
        labels = SemanticSegmentationLabels.make_empty(
            extent=self.extent,
            num_classes=len(self.class_config),
            smooth=False)
        labels[window] = label_arr

        return labels

    def get_label_arr(self, window: Optional[Box] = None) -> np.ndarray:
        """Get labels for a window.

        The returned array will be the same size as the input window. If window
        overflows the extent, the overflowing region will be filled with the ID
        of the null class as defined by the class_config.

        Args:
            window (Optional[Box], optional): Window to get labels for. If
            None, returns a label array covering the full extent of the scene.

        Returns:
            np.ndarray: Label array.
        """
        if window is None:
            window = self.extent

        label_arr = self.raster_source.get_chip(window)
        if label_arr.ndim == 3:
            label_arr = np.squeeze(label_arr, axis=2)
        label_arr = fill_edge(label_arr, window, self.extent,
                              self.class_config.null_class_id)
        return label_arr

    @property
    def extent(self) -> Box:
        return self.raster_source.extent

    def __getitem__(self, key: Any) -> Any:
        if isinstance(key, Box):
            return self.get_label_arr(key)
        else:
            return super().__getitem__(key)



from rastervision.pipeline.config import register_config, Field
from rastervision.core.data.raster_transformer.raster_transformer_config import (  # noqa
    RasterTransformerConfig)
from rastervision.core.data.raster_transformer.cast_transformer import (  # noqa
    CastTransformer)


@register_config('cast_transformer')
class CastTransformerConfig(RasterTransformerConfig):
    """Configure a :class:`.CastTransformer`."""

    to_dtype: str = Field(
        ...,
        description='dtype to cast raster to. Must be a valid Numpy dtype '
        'e.g. "uint8", "float32", etc.')

    def build(self):
        return CastTransformer(to_dtype=self.to_dtype)



from typing import TYPE_CHECKING, Dict, Optional
from copy import deepcopy
import logging

from rastervision.core.data.vector_transformer import VectorTransformer
from rastervision.core.data.vector_transformer.label_maker.filter import (
    create_filter)
from rastervision.core.data.utils.geojson import features_to_geojson

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig, CRSTransformer

log = logging.getLogger(__name__)


class ClassInferenceTransformer(VectorTransformer):
    """Infers missing class_ids from GeoJSON features.

        Rules:
            1) If class_id is in feature['properties'], use it.
            2) If class_config is set and class_name or label are in
                feature['properties'] and in class_config, use corresponding
                class_id.
            3) If class_id_to_filter is set and filter is true when applied to
                feature, use corresponding class_id.
            4) Otherwise, return the default_class_id
    """

    def __init__(self,
                 default_class_id: Optional[int],
                 class_config: Optional['ClassConfig'] = None,
                 class_id_to_filter: Optional[Dict[int, list]] = None):
        self.class_config = class_config
        self.class_id_to_filter = class_id_to_filter
        self.default_class_id = default_class_id

        if self.class_id_to_filter is not None:
            self.class_id_to_filter = {}
            for class_id, filter_exp in class_id_to_filter.items():
                self.class_id_to_filter[int(class_id)] = create_filter(
                    filter_exp)

    @staticmethod
    def infer_feature_class_id(
            feature: dict,
            default_class_id: Optional[int],
            class_config: Optional['ClassConfig'] = None,
            class_id_to_filter: Optional[Dict[int, list]] = None
    ) -> Optional[int]:
        """Infer the class_id for a GeoJSON feature.

        Rules:
            1) If class_id is in feature['properties'], use it.
            2) If class_config is set and class_name or label are in
                feature['properties'] and in class_config, use corresponding
                class_id.
            3) If class_id_to_filter is set and filter is true when applied to
                feature, use corresponding class_id.
            4) Otherwise, return the default_class_id.

        Args:
            feature (dict): GeoJSON feature.

        Returns:
            Optional[int]: Inferred class ID.
        """
        class_id = feature.get('properties', {}).get('class_id')
        if class_id is not None:
            return class_id

        if class_config is not None:
            class_name = feature.get('properties', {}).get('class_name')
            if class_name in class_config.names:
                return class_config.names.index(class_name)

            label = feature.get('properties', {}).get('label')
            if label in class_config.names:
                return class_config.names.index(label)

        if class_id_to_filter is not None:
            for class_id, filter_fn in class_id_to_filter.items():
                if filter_fn(feature):
                    return class_id

        return default_class_id

    def transform(self,
                  geojson: dict,
                  crs_transformer: Optional['CRSTransformer'] = None) -> dict:
        """Add class_id to feature properties and drop features with no class.

        For each feature in geojson, the class_id is inferred and is set into
        feature['properties']. If the class_id is None (because none of the
        rules apply and the default_class_id is None), the feature is dropped.
        """
        new_features = []
        warned = False
        for feature in geojson['features']:
            class_id = self.infer_feature_class_id(
                feature,
                default_class_id=self.default_class_id,
                class_config=self.class_config,
                class_id_to_filter=self.class_id_to_filter)
            if class_id is not None:
                feature = deepcopy(feature)
                properties = feature.get('properties', {})
                properties['class_id'] = class_id
                feature['properties'] = properties
                new_features.append(feature)
            elif not warned:
                log.warning(
                    'ClassInferenceTransformer is dropping vector features because '
                    'class_id cannot be inferred. To avoid this behavior, '
                    'set default_class_id to a non-None value in '
                    'ClassInferenceTransformer.')
                warned = True

        new_geojson = features_to_geojson(new_features)
        return new_geojson



from typing import TYPE_CHECKING, Optional, Sequence, Tuple
from os.path import join
import logging

import numpy as np
import rasterio as rio
from tqdm.auto import tqdm

from rastervision.pipeline.file_system import (
    get_local_path, json_to_file, make_dir, sync_to_dir, file_exists,
    download_if_needed, NotReadableError, get_tmp_dir)
from rastervision.core.box import Box
from rastervision.core.data import (CRSTransformer, ClassConfig)
from rastervision.core.data.label import (SemanticSegmentationLabels,
                                          SemanticSegmentationSmoothLabels)
from rastervision.core.data.label_store import LabelStore
from rastervision.core.data.label_source import SemanticSegmentationLabelSource
from rastervision.core.data.raster_transformer import RGBClassTransformer
from rastervision.core.data.raster_source import RasterioSource

if TYPE_CHECKING:
    from rastervision.core.data import (VectorOutputConfig,
                                        SemanticSegmentationDiscreteLabels)

log = logging.getLogger(__name__)


class SemanticSegmentationLabelStore(LabelStore):
    """Storage for semantic segmentation predictions.

    Can store predicted class ID raster and class scores raster as GeoTIFFs,
    and can optionally vectorize predictions and store them as GeoJSON files.
    """

    def __init__(
            self,
            uri: str,
            extent: Box,
            crs_transformer: CRSTransformer,
            class_config: ClassConfig,
            tmp_dir: Optional[str] = None,
            vector_outputs: Optional[Sequence['VectorOutputConfig']] = None,
            save_as_rgb: bool = False,
            discrete_output: bool = True,
            smooth_output: bool = False,
            smooth_as_uint8: bool = False,
            rasterio_block_size: int = 512):
        """Constructor.

        Args:
            uri (str): Path to directory where the predictions are/will be
                stored. Smooth scores will be saved as "uri/scores.tif",
                discrete labels will be stored as "uri/labels.tif", and vector
                outputs will be saved in "uri/vector_outputs/".
            extent (Box): The extent of the scene.
            crs_transformer (CRSTransformer): CRS transformer for correctly
                mapping from pixel coords to map coords.
            tmp_dir (Optional[str], optional): Temporary directory to use. If
                None, will be auto-generated. Defaults to None.
            vector_outputs (Optional[Sequence[VectorOutputConfig]], optional):
                List of VectorOutputConfig's containing vectorization
                configuration information. Only classes for which a
                VectorOutputConfig is specified will be saved as vectors.
                If None, no vector outputs will be produced. Defaults to None.
            class_config (ClassConfig): Class config.
            save_as_rgb (bool, optional): If True, saves labels as an RGB
                image, using the class-color mapping in the class_config.
                Defaults to False.
            discrete_output (bool, optional): If True, saves labels as a raster
                of class IDs (one band). Defaults to False.
            smooth_output (bool, optional): If True, saves labels as a raster
                of class scores (one band for each class). Defaults to False.
            smooth_as_uint8 (bool, optional): If True, stores smooth class
                scores as np.uint8 (0-255) values rather than as np.float32
                discrete labels, to help save memory/disk space.
                Defaults to False.
            rasterio_block_size (int, optional): Value to set blockxsize and
                blockysize to. Defaults to 512.
        """
        self.root_uri = uri
        self.label_uri = join(uri, 'labels.tif')
        self.score_uri = join(uri, 'scores.tif')
        self.hits_uri = join(uri, 'pixel_hits.npy')

        self.tmp_dir = tmp_dir
        if self.tmp_dir is None:
            self._tmp_dir = get_tmp_dir()
            self.tmp_dir = self._tmp_dir.name

        self.vector_outputs = vector_outputs
        self.extent = extent
        self.crs_transformer = crs_transformer
        self.class_config = class_config
        self.discrete_output = discrete_output
        self.smooth_output = smooth_output
        self.smooth_as_uint8 = smooth_as_uint8
        self.rasterio_block_size = rasterio_block_size

        self.class_transformer = None
        if save_as_rgb:
            self.class_transformer = RGBClassTransformer(class_config)

        self.label_source = None
        self.score_source = None

        if file_exists(self.label_uri):
            if self.class_transformer is not None:
                tfs = [self.class_transformer]
            else:
                tfs = []
            label_raster_source = RasterioSource(
                self.label_uri, raster_transformers=tfs)
            self.label_source = SemanticSegmentationLabelSource(
                label_raster_source, class_config)

        if self.smooth_output:
            if file_exists(self.score_uri):
                raster_source = RasterioSource(self.score_uri)
                extents_equal = raster_source.extent == self.extent
                bands_equal = raster_source.num_channels == len(class_config)
                self_dtype = np.uint8 if self.smooth_as_uint8 else np.float32
                dtypes_equal = raster_source.dtype == self_dtype

                if extents_equal and bands_equal and dtypes_equal:
                    self.score_source = raster_source
                else:
                    raise FileExistsError(f'{self.score_uri} already exists '
                                          'and is incompatible.')

    def get_labels(self) -> SemanticSegmentationLabels:
        """Get all labels.

        Returns:
            SemanticSegmentationLabels
        """
        if self.smooth_output:
            return self.get_scores()
        else:
            return self.get_discrete_labels()

    def get_discrete_labels(self) -> 'SemanticSegmentationDiscreteLabels':
        """Get all labels.

        Returns:
            SemanticSegmentationDiscreteLabels
        """
        if self.label_source is None:
            raise FileNotFoundError(
                f'Raster source at {self.label_uri} does not exist.')

        return self.label_source.get_labels()

    def get_scores(self) -> 'SemanticSegmentationSmoothLabels':
        """Get all scores.

        Returns:
            SemanticSegmentationSmoothLabels
        """
        if self.score_source is None:
            raise Exception(
                f'Raster source at {self.score_uri} does not exist '
                'or is not consistent with the current params.')

        log.info('Loading scores...')

        extent = self.score_source.extent
        try:
            hits_uri_local = download_if_needed(self.hits_uri)
            hits_arr = np.load(hits_uri_local)
        except NotReadableError:
            log.warn(f'Pixel hits array not found at {self.hits_uri}.'
                     'Setting all pixels hits to 1.')
            hits_arr = np.ones(extent.size, dtype=np.uint8)

        score_arr = self.score_source.get_chip(extent)
        # (H, W, C) --> (C, H, W)
        score_arr = score_arr.transpose(2, 0, 1)
        # convert to float
        if score_arr.dtype == np.uint8:
            score_arr = score_arr.astype(np.float16)
            score_arr /= 255

        labels = self.empty_labels()
        assert isinstance(labels, SemanticSegmentationSmoothLabels)
        labels.pixel_scores = score_arr * hits_arr
        labels.pixel_hits = hits_arr
        return labels

    def save(self,
             labels: SemanticSegmentationLabels,
             profile: Optional[dict] = None) -> None:
        """Save labels to disk.

        More info on rasterio IO:
        - https://github.com/mapbox/rasterio/blob/master/docs/quickstart.rst
        - https://rasterio.readthedocs.io/en/latest/topics/windowed-rw.html

        Args:
            labels - (SemanticSegmentationLabels) labels to be saved
        """
        local_root = get_local_path(self.root_uri, self.tmp_dir)
        make_dir(local_root)

        height, width = self.extent.size
        out_profile = dict(
            driver='GTiff',
            height=height,
            width=width,
            transform=self.crs_transformer.transform,
            crs=self.crs_transformer.image_crs,
            blockxsize=min(self.rasterio_block_size, width),
            blockysize=min(self.rasterio_block_size, height))
        if profile is not None:
            out_profile.update(profile)

        # if old scores exist, combine them with the new ones
        if self.score_source is not None:
            log.info('Old scores found. Merging with current scores.')
            old_labels = self.get_scores()
            labels += old_labels

        if self.discrete_output:
            labels_path = get_local_path(self.label_uri, self.tmp_dir)
            self.write_discrete_raster_output(out_profile, labels_path, labels)

        if self.smooth_output:
            scores_path = get_local_path(self.score_uri, self.tmp_dir)
            hits_path = get_local_path(self.hits_uri, self.tmp_dir)
            self.write_smooth_raster_output(out_profile, scores_path,
                                            hits_path, labels)

        if self.vector_outputs is not None:
            self.write_vector_outputs(labels)

        sync_to_dir(local_root, self.root_uri)

    def write_smooth_raster_output(
            self, out_profile: dict, scores_path: str, hits_path: str,
            labels: SemanticSegmentationSmoothLabels) -> None:
        num_bands = labels.num_classes
        dtype = np.uint8 if self.smooth_as_uint8 else np.float32
        out_profile.update(dict(count=num_bands, dtype=dtype))

        with rio.open(scores_path, 'w', **out_profile) as ds:
            windows = [Box.from_rasterio(w) for _, w in ds.block_windows(1)]
            with tqdm(windows, desc='Saving pixel scores') as bar:
                for window in bar:
                    window, _ = self._clip_to_extent(self.extent, window)
                    score_arr = labels.get_score_arr(window)
                    if dtype == np.uint8:
                        score_arr = self._scores_to_uint8(score_arr)
                    else:
                        score_arr = score_arr.astype(dtype)
                    self._write_array(ds, window, score_arr)

        # save pixel hits too
        np.save(hits_path, labels.pixel_hits)

    def write_discrete_raster_output(
            self, out_profile: dict, path: str,
            labels: SemanticSegmentationLabels) -> None:
        num_bands = 1 if self.class_transformer is None else 3
        dtype = np.uint8
        out_profile.update(dict(count=num_bands, dtype=dtype))

        null_class_id = self.class_config.null_class_id
        with rio.open(path, 'w', **out_profile) as ds:
            windows = [Box.from_rasterio(w) for _, w in ds.block_windows(1)]
            with tqdm(windows, desc='Saving pixel labels') as bar:
                for window in bar:
                    label_arr = labels.get_label_arr(
                        window, null_class_id).astype(dtype)
                    window, label_arr = self._clip_to_extent(
                        self.extent, window, label_arr)
                    if self.class_transformer is not None:
                        label_arr = self.class_transformer.class_to_rgb(
                            label_arr)
                        label_arr = label_arr.transpose(2, 0, 1)
                    self._write_array(ds, window, label_arr)

    def write_vector_outputs(self, labels: SemanticSegmentationLabels) -> None:
        """Write vectorized outputs for all configs in self.vector_outputs."""
        from rastervision.core.data.utils import (denoise, geoms_to_geojson,
                                                  mask_to_building_polygons,
                                                  mask_to_polygons)

        log.info('Writing vector output to disk.')

        label_arr = labels.get_label_arr(self.extent,
                                         self.class_config.null_class_id)

        with tqdm(self.vector_outputs, desc='Vectorizing predictions') as bar:
            for i, vo in enumerate(bar):
                bar.set_postfix(
                    dict(
                        class_id=vo.class_id,
                        mode=vo.get_mode(),
                        denoise_radius=vo.denoise))

                if vo.uri is None:
                    log.info(f'Skipping VectorOutputConfig at index {i} '
                             'due to missing uri.')
                    continue

                class_mask = (label_arr == vo.class_id).astype(np.uint8)

                if vo.denoise > 0:
                    class_mask = denoise(class_mask, radius=vo.denoise)

                mode = vo.get_mode()
                if mode == 'polygons':
                    polys = mask_to_polygons(class_mask)
                elif mode == 'buildings':
                    polys = mask_to_building_polygons(
                        mask=class_mask,
                        min_area=vo.min_area,
                        width_factor=vo.element_width_factor,
                        thickness=vo.element_thickness)
                else:
                    raise NotImplementedError()

                polys = [self.crs_transformer.pixel_to_map(p) for p in polys]
                geojson = geoms_to_geojson(polys)
                json_to_file(geojson, vo.uri)

    def empty_labels(self, **kwargs) -> SemanticSegmentationLabels:
        """Returns an empty SemanticSegmentationLabels object."""
        args = dict(
            extent=self.extent,
            num_classes=len(self.class_config),
            smooth=self.smooth_output)
        args.update(**kwargs)
        labels = SemanticSegmentationLabels.make_empty(**args)
        return labels

    def _write_array(self, dataset: rio.DatasetReader, window: Box,
                     arr: np.ndarray) -> None:
        """Write array out to a rasterio dataset. Array must be of shape
        (C, H, W).
        """
        rio_window = window.rasterio_format()
        if len(arr.shape) == 2:
            dataset.write_band(1, arr, window=rio_window)
        else:
            for i, band in enumerate(arr, start=1):
                dataset.write_band(i, band, window=rio_window)

    def _clip_to_extent(self,
                        extent: Box,
                        window: Box,
                        arr: Optional[np.ndarray] = None
                        ) -> Tuple[Box, Optional[np.ndarray]]:
        clipped_window = window.intersection(extent)
        if arr is not None:
            h, w = clipped_window.size
            arr = arr[:h, :w]
        return clipped_window, arr

    def _scores_to_uint8(self, score_arr: np.ndarray) -> np.ndarray:
        """Quantize scores to uint8 (0-255)."""
        score_arr *= 255
        score_arr = np.around(score_arr, out=score_arr)
        score_arr = score_arr.astype(np.uint8)
        return score_arr



# flake8: noqa

from rastervision.core.data.class_config import *
from rastervision.core.data.raster_source import *
from rastervision.core.data.crs_transformer import *
from rastervision.core.data.label import *
from rastervision.core.data.vector_source import *
from rastervision.core.data.label_source import *
from rastervision.core.data.label_store import *
from rastervision.core.data.scene import *
from rastervision.core.data.scene_config import *
from rastervision.core.data.dataset_config import *
from rastervision.core.data.raster_transformer import *
from rastervision.core.data.vector_transformer import *
from rastervision.core.data.utils import *



from typing import TYPE_CHECKING, Any, List, Optional, Sequence, Tuple, Union
import logging
import os
import subprocess

import numpy as np
import rasterio
from rasterio.enums import (ColorInterp, MaskFlags, Resampling)

from rastervision.pipeline.file_system import download_if_needed, get_tmp_dir
from rastervision.core.box import Box
from rastervision.core.data.crs_transformer import RasterioCRSTransformer
from rastervision.core.data.raster_source import RasterSource
from rastervision.core.data.utils import listify_uris

if TYPE_CHECKING:
    from rasterio.io import DatasetReader
    from rastervision.core.data import RasterTransformer

log = logging.getLogger(__name__)


def build_vrt(vrt_path: str, image_uris: List[str]) -> None:
    """Build a VRT for a set of TIFF files.

    Args:
        vrt_path (str): Local path for the VRT to be created.
        image_uris (List[str]): Image URIs.
    """
    log.info('Building VRT...')
    cmd = ['gdalbuildvrt', vrt_path]
    cmd.extend(image_uris)
    subprocess.run(cmd)


def download_and_build_vrt(image_uris: List[str],
                           vrt_dir: str,
                           stream: bool = False) -> str:
    """Download images (if needed) and build a VRT for a set of TIFF files.

    Args:
        image_uris (List[str]): Image URIs.
        vrt_dir (str): Dir where the VRT will be created.
        stream (bool, optional): If true, do not download images.
            Defaults to False.

    Returns:
        str: The path to the created VRT file.
    """
    if not stream:
        image_uris = [download_if_needed(uri) for uri in image_uris]
    vrt_path = os.path.join(vrt_dir, 'index.vrt')
    build_vrt(vrt_path, image_uris)
    return vrt_path


def load_window(
        image_dataset: 'DatasetReader',
        bands: Optional[Union[int, Sequence[int]]] = None,
        window: Optional[Tuple[Tuple[int, int], Tuple[int, int]]] = None,
        is_masked: bool = False,
        out_shape: Optional[Tuple[int, ...]] = None) -> np.ndarray:
    """Load a window of an image using Rasterio.

    Args:
        image_dataset: a Rasterio dataset.
        bands (Optional[Union[int, Sequence[int]]]): Band index or indices to
            read. Must be 1-indexed.
        window (Optional[Tuple[Tuple[int, int], Tuple[int, int]]]):
            ((row_start, row_stop), (col_start, col_stop)) or
            ((y_min, y_max), (x_min, x_max)). If None, reads the entire raster.
            Defaults to None.
        is_masked (bool): If True, read a masked array from rasterio.
            Defaults to False.
        out_shape (Optional[Tuple[int, int]]): (hieght, width) of the output
            chip. If None, no resizing is done. Defaults to None.

    Returns:
        np.ndarray: array of shape (height, width, channels).
    """
    if bands is not None:
        bands = tuple(bands)
    im = image_dataset.read(
        indexes=bands,
        window=window,
        boundless=True,
        masked=is_masked,
        out_shape=out_shape,
        resampling=Resampling.bilinear)

    if is_masked:
        im = np.ma.filled(im, fill_value=0)

    # Handle non-zero NODATA values by setting the data to 0.
    if bands is None:
        for channel, nodataval in enumerate(image_dataset.nodatavals):
            if nodataval is not None and nodataval != 0:
                im[channel, im[channel] == nodataval] = 0
    else:
        for channel, src_band in enumerate(bands):
            src_band_0_indexed = src_band - 1
            nodataval = image_dataset.nodatavals[src_band_0_indexed]
            if nodataval is not None and nodataval != 0:
                im[channel, im[channel] == nodataval] = 0

    im = np.transpose(im, axes=[1, 2, 0])
    return im


def fill_overflow(extent: Box,
                  window: Box,
                  chip: np.ndarray,
                  fill_value: int = 0) -> np.ndarray:
    """Where ``chip``'s ``window`` overflows extent, fill with ``fill_value``.

    Args:
        extent (Box): Extent.
        window (Box): Window from which ``chip`` was read.
        chip (np.ndarray): (H, W, C) array.
        fill_value (int, optional): Value to set oveflowing pixels to.
            Defaults to 0.

    Returns:
        np.ndarray: Chip with overflowing regions filled with ``fill_value``.
    """
    top_overflow = max(0, extent.ymin - window.ymin)
    bottom_overflow = max(0, window.ymax - extent.ymax)
    left_overflow = max(0, extent.xmin - window.xmin)
    right_overflow = max(0, window.xmax - extent.xmax)

    h, w = chip.shape[:2]
    chip[:top_overflow] = fill_value
    chip[h - bottom_overflow:] = fill_value
    chip[:, :left_overflow] = fill_value
    chip[:, w - right_overflow:] = fill_value
    return chip


def get_channel_order_from_dataset(
        image_dataset: 'DatasetReader') -> List[int]:
    """Get channel order from rasterio image dataset.

    Accounts for dataset's ``colorinterp`` if defined.

    Args:
        image_dataset (DatasetReader): Rasterio image dataset.

    Returns:
        List[int]: List of channel indices.
    """
    colorinterp = image_dataset.colorinterp
    if colorinterp:
        channel_order = [
            i for i, color_interp in enumerate(colorinterp)
            if color_interp != ColorInterp.alpha
        ]
    else:
        channel_order = list(range(0, image_dataset.count))
    return channel_order


class RasterioSource(RasterSource):
    """A rasterio-based :class:`.RasterSource`.

    This RasterSource can read any file that can be opened by
    `Rasterio/GDAL <https://www.gdal.org/formats_list.html>`_.

    If there are multiple image files that cover a single scene, you can pass
    the corresponding list of URIs, and read them as if it were a single
    stitched-together image.

    It can also read non-georeferenced images such as ``.tif``, ``.png``, and
    ``.jpg`` files. This is useful for oblique drone imagery, biomedical
    imagery, and any other (potentially massive!) non-georeferenced images.

    If channel_order is None, then use non-alpha channels. This also sets any
    masked or NODATA pixel values to be zeros.
    """

    def __init__(self,
                 uris: Union[str, List[str]],
                 raster_transformers: List['RasterTransformer'] = [],
                 allow_streaming: bool = False,
                 channel_order: Optional[Sequence[int]] = None,
                 extent: Optional[Box] = None,
                 tmp_dir: Optional[str] = None):
        """Constructor.

        Args:
            uris (Union[str, List[str]]): One or more URIs of images. If more
                than one, the images will be mosaiced together using GDAL.
            raster_transformers (List['RasterTransformer']): RasterTransformers
                to use to trasnform chips after they are read.
            allow_streaming (bool): If True, read data without downloading the
                entire file first. Defaults to False.
            channel_order (Optional[Sequence[int]]): List of indices of
                channels to extract from raw imagery. Can be a subset of the
                available channels. If None, all channels available in the
                image will be read. Defaults to None.
            extent (Optional[Box], optional): Use-specified extent. If None,
                the full extent of the raster source is used.
            tmp_dir (Optional[str]): Directory to use for storing the VRT
                (needed if multiple uris or allow_streaming=True). If None,
                will be auto-generated. Defaults to None.
        """
        self.uris = listify_uris(uris)
        self.allow_streaming = allow_streaming
        self._num_channels = None
        self._dtype = None

        self.tmp_dir = tmp_dir
        if self.tmp_dir is None:
            self._tmp_dir = get_tmp_dir()
            self.tmp_dir = self._tmp_dir.name

        self.imagery_path = self.download_data(
            self.tmp_dir, stream=self.allow_streaming)
        self.image_dataset = rasterio.open(self.imagery_path)

        block_shapes = set(self.image_dataset.block_shapes)
        if len(block_shapes) > 1:
            log.warn('Raster bands have non-identical block shapes: '
                     f'{block_shapes}. This can slow down reading. '
                     'Consider re-tiling using GDAL.')

        for h, w in block_shapes:
            # the choice of 4 here is arbitrary
            if max(h, w) / min(h, w) > 4:
                log.warn(f'Raster block size {(h, w)} is too non-square. '
                         'This can slow down reading. '
                         'Consider re-tiling using GDAL.')

        self._crs_transformer = RasterioCRSTransformer.from_dataset(
            self.image_dataset)

        num_channels_raw = self.image_dataset.count
        if channel_order is None:
            channel_order = get_channel_order_from_dataset(self.image_dataset)
        self.bands_to_read = np.array(channel_order, dtype=int) + 1

        # number of output channels
        if len(raster_transformers) == 0:
            self._num_channels = len(self.bands_to_read)

        mask_flags = self.image_dataset.mask_flag_enums
        self.is_masked = any(m for m in mask_flags if m != MaskFlags.all_valid)

        if extent is None:
            height = self.image_dataset.height
            width = self.image_dataset.width
            extent = Box(0, 0, height, width)

        super().__init__(
            channel_order,
            num_channels_raw,
            raster_transformers=raster_transformers,
            extent=extent)

    @property
    def num_channels(self) -> int:
        """Number of channels in the chips read from this source.

        .. note::

            Unlike the parent class, ``RasterioSource`` applies channel_order
            (via ``bands_to_read``) before ``raster_transformers``. So the
            number of output channels is not guaranteed to be equal to
            ``len(channel_order)``.
        """
        if self._num_channels is None:
            self._set_info_from_chip()
        return self._num_channels

    @property
    def dtype(self) -> Tuple[int, int, int]:
        if self._dtype is None:
            self._set_info_from_chip()
        return self._dtype

    @property
    def crs_transformer(self) -> RasterioCRSTransformer:
        return self._crs_transformer

    def _set_info_from_chip(self):
        """Read 1x1 chip to get info not statically inferrable."""
        test_chip = self.get_chip(Box(0, 0, 1, 1))
        self._dtype = test_chip.dtype
        self._num_channels = test_chip.shape[-1]

    def download_data(self,
                      vrt_dir: Optional[str] = None,
                      stream: bool = False) -> str:
        """Download any data needed for this raster source.

        Return a single local path representing the image or a VRT of the data.
        """
        if len(self.uris) == 1:
            if stream:
                return self.uris[0]
            else:
                return download_if_needed(self.uris[0])
        else:
            if vrt_dir is None:
                raise ValueError('vrt_dir is required if using >1 image URIs.')
            return download_and_build_vrt(self.uris, vrt_dir, stream=stream)

    def _get_chip(self,
                  window: Box,
                  bands: Optional[Sequence[int]] = None,
                  out_shape: Optional[Tuple[int, ...]] = None) -> np.ndarray:
        window = window.shift_origin(self.extent)
        chip = load_window(
            self.image_dataset,
            bands=bands,
            window=window.rasterio_format(),
            is_masked=self.is_masked,
            out_shape=out_shape)
        chip = fill_overflow(self.extent, window, chip)
        return chip

    def get_chip(self,
                 window: Box,
                 bands: Optional[Union[Sequence[int], slice]] = None,
                 out_shape: Optional[Tuple[int, ...]] = None) -> np.ndarray:
        """Read a chip specified by a window from the file.

        Args:
            window (Box): Bounding box of chip in pixel coordinates.
            bands (Optional[Union[Sequence[int], slice]], optional): Subset of
                bands to read. Note that this will be applied on top of the
                channel_order (if specified). So if this is an RGB image and
                channel_order=[2, 1, 0], then using bands=[0] will return the
                B-channel. Defaults to None.
            out_shape (Optional[Tuple[int, ...]], optional): (hieght, width) of
            the output chip. If None, no resizing is done. Defaults to None.

        Returns:
            np.ndarray: A chip of shape (height, width, channels).
        """
        bands_to_read = self.bands_to_read
        if bands is not None:
            bands_to_read = bands_to_read[bands]
        chip = self._get_chip(window, out_shape=out_shape, bands=bands_to_read)
        for transformer in self.raster_transformers:
            chip = transformer.transform(chip, self.channel_order)
        return chip

    def __getitem__(self, key: Any) -> 'np.ndarray':
        if isinstance(key, Box):
            return self.get_chip(key)
        elif isinstance(key, slice):
            key = [key]
        elif isinstance(key, tuple):
            pass
        else:
            raise TypeError('Unsupported key type.')

        slices = list(key)
        assert 1 <= len(slices) <= 3
        assert all(s is not None for s in slices)
        assert isinstance(slices[0], slice)
        if len(slices) == 1:
            h, = slices
            w = slice(None, None)
            c = None
        elif len(slices) == 2:
            assert isinstance(slices[1], slice)
            h, w = slices
            c = None
        else:
            h, w, c = slices

        if any(x is not None and x < 0
               for x in [h.start, h.stop, w.start, w.stop]):
            raise NotImplementedError()

        ymin, xmin, ymax, xmax = self.extent
        _ymin = 0 if h.start is None else h.start
        _xmin = 0 if w.start is None else w.start
        _ymax = ymax if h.stop is None else h.stop
        _xmax = xmax if w.stop is None else w.stop
        window = Box(_ymin, _xmin, _ymax, _xmax)

        out_shape = None
        if h.step is not None or w.step is not None:
            out_h, out_w = window.size
            if h.step is not None:
                out_h //= h.step
            if w.step is not None:
                out_w //= w.step
            out_shape = (int(out_h), int(out_w))

        chip = self.get_chip(window, bands=c, out_shape=out_shape)
        return chip



from rastervision.core.evaluation import (ClassificationEvaluator,
                                          ObjectDetectionEvaluation)


class ObjectDetectionEvaluator(ClassificationEvaluator):
    """Evaluates predictions for a set of scenes.
    """

    def __init__(self, class_config, output_uri):
        super().__init__(class_config, output_uri)

    def create_evaluation(self):
        return ObjectDetectionEvaluation(self.class_config)



import logging
from os.path import join
import tempfile
import shutil
from typing import TYPE_CHECKING, Optional, List
from functools import lru_cache

import click
import numpy as np
from tqdm.auto import tqdm

from rastervision.pipeline.pipeline import Pipeline
from rastervision.core.box import Box
from rastervision.core.data_sample import DataSample
from rastervision.core.data import Scene, Labels
from rastervision.core.backend import Backend
from rastervision.core.rv_pipeline import TRAIN, VALIDATION
from rastervision.pipeline.file_system.utils import (
    download_if_needed, zipdir, get_local_path, upload_or_copy, make_dir,
    sync_to_dir, file_exists)

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from rastervision.core.rv_pipeline import RVPipelineConfig

ALL_COMMANDS = ['analyze', 'chip', 'train', 'predict', 'eval', 'bundle']
SPLITTABLE_COMMANDS = ['chip', 'predict']
GPU_COMMANDS = ['train', 'predict']


class RVPipeline(Pipeline):
    """Base class of all Raster Vision Pipelines.

    This can be subclassed to implement Pipelines for different computer vision
    tasks over geospatial imagery. The commands and what they produce include:

    - analyze: metrics on the imagery and labels
    - chip: small training and validation images taken from larger scenes
    - train: model trained on chips
    - predict: predictions over entire validation and test scenes
    - eval: evaluation metrics for predictions generated by model
    - bundle: bundle containing model and any other files needed to make

    predictions using the Predictor.
    """

    def __init__(self, config: 'RVPipelineConfig', tmp_dir: str):
        super().__init__(config, tmp_dir)
        self.backend: Optional['Backend'] = None
        self.config: 'RVPipelineConfig'

    @property
    def commands(self):
        commands = ALL_COMMANDS[:]
        if len(self.config.analyzers) == 0 and 'analyze' in commands:
            commands.remove('analyze')
            click.secho("Skipping 'analyze' command...", fg='green', bold=True)
        commands = self.config.backend.filter_commands(commands)
        return commands

    @property
    def split_commands(self):
        return self.config.backend.filter_commands(SPLITTABLE_COMMANDS)

    @property
    def gpu_commands(self):
        return self.config.backend.filter_commands(GPU_COMMANDS)

    def analyze(self):
        """Run each analyzer over training scenes."""
        dataset = self.config.dataset
        class_config = dataset.class_config

        scene_id_to_cfg = {s.id: s for s in dataset.all_scenes}

        @lru_cache(maxsize=len(dataset.all_scenes))
        def build_scene(scene_id: str) -> Scene:
            cfg = scene_id_to_cfg[scene_id]
            scene = cfg.build(
                class_config, self.tmp_dir, use_transformers=False)
            return scene

        # build and run each AnalyzerConfig for each scene group
        for a in self.config.analyzers:
            for group_name, group_ids in dataset.scene_groups.items():
                if len(group_ids) == 0:
                    log.info(f'Skipping scene group "{group_name}". '
                             'Empty scene group.')
                    continue
                group_scenes = (build_scene(id) for id in group_ids)
                analyzer = a.build(scene_group=(group_name, group_scenes))

                log.info(f'Running {type(analyzer).__name__} on '
                         f'scene group "{group_name}"...')
                analyzer.process(group_scenes, self.tmp_dir)

    def get_train_windows(self, scene: Scene) -> List[Box]:
        """Return the training windows for a Scene.

        Each training window represents the spatial extent of a training chip to
        generate.

        Args:
            scene: Scene to generate windows for
        """
        raise NotImplementedError()

    def get_train_labels(self, window: Box, scene: Scene) -> Labels:
        """Return the training labels in a window for a scene.

        Returns:
            Labels that lie within window
        """
        raise NotImplementedError()

    def chip(self, split_ind: int = 0, num_splits: int = 1):
        """Save training and validation chips."""
        cfg = self.config
        backend = cfg.backend.build(cfg, self.tmp_dir)
        dataset = cfg.dataset.get_split_config(split_ind, num_splits)
        if not dataset.train_scenes and not dataset.validation_scenes:
            return

        class_cfg = dataset.class_config
        with backend.get_sample_writer() as writer:

            def chip_scene(scene, split):
                windows = self.get_train_windows(scene)
                with tqdm(
                        windows,
                        desc=f'Making {split} chips from scene {scene.id}',
                        mininterval=0.5) as bar:
                    for window in bar:
                        chip = scene.raster_source.get_chip(window)
                        labels = self.get_train_labels(window, scene)
                        sample = DataSample(
                            chip=chip,
                            window=window,
                            labels=labels,
                            scene_id=str(scene.id),
                            is_train=split == TRAIN)
                        sample = self.post_process_sample(sample)
                        writer.write_sample(sample)

            for s in dataset.train_scenes:
                chip_scene(s.build(class_cfg, self.tmp_dir), TRAIN)
            for s in dataset.validation_scenes:
                chip_scene(s.build(class_cfg, self.tmp_dir), VALIDATION)

    def train(self):
        """Train a model and save it."""
        backend = self.config.backend.build(self.config, self.tmp_dir)
        backend.train(source_bundle_uri=self.config.source_bundle_uri)

    def post_process_sample(self, sample: DataSample) -> DataSample:
        """Post-process sample in pipeline-specific way.

        This should be called before writing a sample during chipping.
        """
        return sample

    def post_process_batch(self, windows: List[Box], chips: np.ndarray,
                           labels: Labels) -> Labels:
        """Post-process a batch of predictions."""
        return labels

    def post_process_predictions(self, labels: Labels, scene: Scene) -> Labels:
        """Post-process all labels at end of prediction."""
        return labels

    def predict(self, split_ind=0, num_splits=1):
        """Make predictions over each validation and test scene.

        This uses a sliding window.
        """
        # Cache backend so subsquent calls will be faster. This is useful for
        # the predictor.
        if self.backend is None:
            self.backend = self.config.backend.build(self.config, self.tmp_dir)
            self.backend.load_model()

        class_config = self.config.dataset.class_config
        dataset = self.config.dataset.get_split_config(split_ind, num_splits)

        for scene_config in (dataset.validation_scenes + dataset.test_scenes):
            scene = scene_config.build(class_config, self.tmp_dir)
            labels = self.predict_scene(scene, self.backend)
            labels = self.post_process_predictions(labels, scene)
            scene.label_store.save(labels)

    def predict_scene(self, scene: Scene, backend: Backend) -> Labels:
        chip_sz = self.config.predict_chip_sz
        stride = chip_sz
        return backend.predict_scene(scene, chip_sz=chip_sz, stride=stride)

    def eval(self):
        """Evaluate predictions against ground truth."""
        dataset = self.config.dataset
        class_config = dataset.class_config
        # it might make sense to make excluded_groups a field in an EvalConfig
        # in the future
        excluded_groups = ['train_scenes']

        scene_id_to_cfg = {s.id: s for s in dataset.all_scenes}

        @lru_cache(maxsize=len(dataset.all_scenes))
        def build_scene(scene_id: str) -> Scene:
            cfg = scene_id_to_cfg[scene_id]
            scene = cfg.build(
                class_config, self.tmp_dir, use_transformers=True)
            return scene

        # build and run each EvaluatorConfig for each scene group
        for e in self.config.evaluators:
            for group_name, group_ids in dataset.scene_groups.items():
                if group_name in excluded_groups:
                    continue
                if len(group_ids) == 0:
                    log.info(f'Skipping scene group "{group_name}". '
                             'Empty scene group.')
                    continue
                group_scenes = (build_scene(id) for id in group_ids)
                evaluator = e.build(
                    class_config, scene_group=(group_name, group_scenes))

                log.info(f'Running {type(evaluator).__name__} on '
                         f'scene group "{group_name}"...')
                try:
                    evaluator.process(group_scenes, self.tmp_dir)
                except FileNotFoundError:
                    log.warn(f'Skipping scene group "{group_name}". '
                             'Either labels or predictions are missing for '
                             'some scene.')

    def bundle(self):
        """Save a model bundle with whatever is needed to make predictions.

        The model bundle is a zip file and it is used by the Predictor and
        predict CLI subcommand.
        """
        with tempfile.TemporaryDirectory(dir=self.tmp_dir) as tmp_dir:
            bundle_dir = join(tmp_dir, 'bundle')
            make_dir(bundle_dir)

            for fn in self.config.backend.get_bundle_filenames():
                path = download_if_needed(
                    join(self.config.train_uri, fn), tmp_dir)
                shutil.copy(path, join(bundle_dir, fn))

            if file_exists(self.config.analyze_uri, include_dir=True):
                sync_to_dir(self.config.analyze_uri, join(
                    bundle_dir, 'analyze'))

            path = download_if_needed(self.config.get_config_uri(), tmp_dir)
            shutil.copy(path, join(bundle_dir, 'pipeline-config.json'))

            model_bundle_uri = self.config.get_model_bundle_uri()
            model_bundle_path = get_local_path(model_bundle_uri, self.tmp_dir)
            zipdir(bundle_dir, model_bundle_path)
            upload_or_copy(model_bundle_path, model_bundle_uri)



from typing import TYPE_CHECKING, Any, List, Optional, Tuple
from abc import ABC, abstractmethod, abstractproperty

from rastervision.core.box import Box

if TYPE_CHECKING:
    from rastervision.core.data import (CRSTransformer, RasterTransformer)
    import numpy as np


class ChannelOrderError(Exception):
    def __init__(self, channel_order: List[int], num_channels_raw: int):
        self.channel_order = channel_order
        self.num_channels_raw = num_channels_raw
        msg = (f'The channel_order ({channel_order}) contains an'
               f'index >= num_channels_raw ({num_channels_raw}).')
        super().__init__(msg)


class RasterSource(ABC):
    """A source of raster data.

    This should be subclassed when adding a new source of raster data such as
    a set of files, an API, a TMS URI schema, etc.
    """

    def __init__(self,
                 channel_order: Optional[List[int]],
                 num_channels_raw: int,
                 raster_transformers: List['RasterTransformer'] = [],
                 extent: Optional[Box] = None):
        """Constructor.

        Args:
            channel_order: list of channel indices to use when extracting chip
                from raw imagery.
            num_channels_raw: Number of channels in the raw imagery before
                applying channel_order.
            raster_transformers: ``RasterTransformers`` for transforming chips
                whenever they are retrieved. Defaults to ``[]``.
            extent: Use-specified extent. If None, the full extent of the
                raster source is used.
        """
        if channel_order is None:
            channel_order = list(range(num_channels_raw))

        if any(c >= num_channels_raw for c in channel_order):
            raise ChannelOrderError(channel_order, num_channels_raw)

        self.channel_order = channel_order
        self.num_channels_raw = num_channels_raw
        self.raster_transformers = raster_transformers
        self._extent = extent

    @property
    def num_channels(self) -> int:
        """Number of channels in the chips read from this source."""
        return len(self.channel_order)

    @property
    def shape(self) -> Tuple[int, int, int]:
        """Shape of the raster as a (height, width, num_channels) tuple."""
        ymin, xmin, ymax, xmax = self.extent
        return ymax - ymin, xmax - xmin, self.num_channels

    @abstractproperty
    def dtype(self) -> 'np.dtype':
        """``numpy.dtype`` of the chips read from this source."""
        pass

    @property
    def extent(self) -> 'Box':
        """Extent of the RasterSource."""
        return self._extent

    @abstractproperty
    def crs_transformer(self) -> 'CRSTransformer':
        """Associated :class:`.CRSTransformer`."""
        pass

    @abstractmethod
    def _get_chip(self, window: 'Box') -> 'np.ndarray':
        """Return raw chip without applying channel_order or transforms.

        Args:
            window: Box

        Returns:
            [height, width, channels] numpy array
        """
        pass

    def __getitem__(self, key: Any) -> 'np.ndarray':
        if isinstance(key, Box):
            return self.get_chip(key)
        elif isinstance(key, slice):
            key = [key]
        elif isinstance(key, tuple):
            pass
        else:
            raise TypeError('Unsupported key type.')
        slices = list(key)

        assert 1 <= len(slices) <= 2
        assert all(s is not None for s in slices)
        assert isinstance(slices[0], slice)
        if len(slices) == 1:
            h, = slices
            w = slice(None, None)
        else:
            assert isinstance(slices[1], slice)
            h, w = slices

        if any(x is not None and x < 0
               for x in [h.start, h.stop, w.start, w.stop]):
            raise NotImplementedError()

        ymin, xmin, ymax, xmax = self.extent
        _ymin = 0 if h.start is None else h.start
        _xmin = 0 if w.start is None else w.start
        _ymax = ymax if h.stop is None else h.stop
        _xmax = xmax if w.stop is None else w.stop
        window = Box(_ymin, _xmin, _ymax, _xmax)

        chip = self.get_chip(window)
        if h.step is not None or w.step is not None:
            chip = chip[::h.step, ::w.step]
        return chip

    def get_chip(self, window: 'Box') -> 'np.ndarray':
        """Return the transformed chip in the window.

        Get a raw chip, extract subset of channels using channel_order, and then apply
        transformations.

        Args:
            window (Box): The window for which to get the chip.

        Returns:
            np.ndarray: Array of shape (height, width, channels).
        """
        chip = self._get_chip(window)

        chip = chip[:, :, self.channel_order]

        for transformer in self.raster_transformers:
            chip = transformer.transform(chip, self.channel_order)

        return chip

    def get_raw_chip(self, window: 'Box') -> 'np.ndarray':
        """Return raw chip without applying channel_order or transforms.

        Args:
            window (Box): The window for which to get the chip.

        Returns:
            np.ndarray: Array of shape (height, width, channels).
        """
        return self._get_chip(window)

    def get_image_array(self) -> 'np.ndarray':
        """Return entire transformed image array.

        .. warning:: Not safe to call on very large RasterSources.

        Returns:
            np.ndarray: Array of shape (height, width, channels).
        """
        return self.get_chip(self.extent)

    def get_raw_image_array(self) -> 'np.ndarray':
        """Return raw image for the full extent.

        .. warning:: Not safe to call on very large RasterSources.

        Returns:
            np.ndarray: Array of shape (height, width, channels).
        """
        return self.get_raw_chip(self.extent)



from typing import TYPE_CHECKING
import logging

from rastervision.core.evaluation import (ClassificationEvaluator,
                                          SemanticSegmentationEvaluation)

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from rastervision.core.data import Scene


class SemanticSegmentationEvaluator(ClassificationEvaluator):
    """Evaluates predictions for a set of scenes."""

    def create_evaluation(self) -> SemanticSegmentationEvaluation:
        return SemanticSegmentationEvaluation(self.class_config)

    def evaluate_scene(self, scene: 'Scene') -> SemanticSegmentationEvaluation:
        """Override to pass null_class_id to filter_by_aoi()."""
        null_class_id = self.class_config.null_class_id
        ground_truth = scene.label_source.get_labels()
        predictions = scene.label_store.get_labels()

        if scene.aoi_polygons:
            ground_truth = ground_truth.filter_by_aoi(scene.aoi_polygons,
                                                      null_class_id)
            predictions = predictions.filter_by_aoi(scene.aoi_polygons,
                                                    null_class_id)
        evaluation = self.evaluate_predictions(ground_truth, predictions)
        return evaluation



from typing import Optional

from rastervision.core.data.raster_transformer.raster_transformer \
    import RasterTransformer

import numpy as np


class CastTransformer(RasterTransformer):
    """Casts chips to the specified dtype."""

    def __init__(self, to_dtype: str):
        """Constructor.

        Args:
            to_dtype: (str) dtype to cast the chips to.
        """
        self.to_dtype = np.dtype(to_dtype)

    def __repr__(self):
        return f'CastTransformer(to_dtype="{self.to_dtype}")'

    def transform(self, chip: np.ndarray,
                  channel_order: Optional[list] = None) -> np.ndarray:
        """Cast chip to self.to_dtype.

        Args:
            chip: ndarray of shape [height, width, channels]

        Returns:
            [height, width, channels] numpy array
        """
        return chip.astype(self.to_dtype)



from typing import TYPE_CHECKING, Dict, Optional

from rastervision.pipeline.config import register_config, Field
from rastervision.core.data.vector_transformer import (VectorTransformerConfig,
                                                       BufferTransformer)

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig


@register_config('buffer_transformer')
class BufferTransformerConfig(VectorTransformerConfig):
    """Configure a :class:`.BufferTransformer`.

    This is useful, for example, for buffering lines representing roads so that
    their width roughly matches the width of roads in the imagery.
    """
    geom_type: str = Field(
        ...,
        description='The geometry type to apply this transform to. '
        'E.g. "LineString", "Point", "Polygon".')
    class_bufs: Dict[int, Optional[float]] = Field(
        {},
        description='Mapping from class IDs to buffer amounts (in pixels). '
        'If a class ID is not found in the mapping, the value specified by '
        'the default_buf field will be used. If the buffer value for a class '
        'is None, then no buffering will be applied to the geoms of that '
        'class and the geom won\'t get converted to a Polygon. Not converting '
        'to Polygon is incompatible with the currently available '
        'LabelSources, but may be useful in the future.')
    default_buf: Optional[float] = Field(
        1,
        description='Default buffer to apply to classes not in class_bufs. '
        'If None, no buffering will be applied to the geoms of those classes.')

    def build(self, class_config: Optional['ClassConfig'] = None
              ) -> BufferTransformer:
        return BufferTransformer(
            self.geom_type,
            class_bufs=self.class_bufs,
            default_buf=self.default_buf)



from rastervision.pipeline.config import register_config
from rastervision.core.data.raster_transformer import (RasterTransformerConfig,
                                                       MinMaxTransformer)


@register_config('min_max_transformer')
class MinMaxTransformerConfig(RasterTransformerConfig):
    """Configure a :class:`.MinMaxTransformer`."""

    def build(self):
        return MinMaxTransformer()



from rastervision.pipeline.config import register_config
from rastervision.core.rv_pipeline import RVPipelineConfig
from rastervision.core.data.label_store import (
    ChipClassificationGeoJSONStoreConfig)
from rastervision.core.evaluation import ChipClassificationEvaluatorConfig


@register_config('chip_classification')
class ChipClassificationConfig(RVPipelineConfig):
    """Configure a :class:`.ChipClassification` pipeline."""

    def build(self, tmp_dir):
        from rastervision.core.rv_pipeline.chip_classification import ChipClassification
        return ChipClassification(self, tmp_dir)

    def get_default_label_store(self, scene):
        return ChipClassificationGeoJSONStoreConfig()

    def get_default_evaluator(self):
        return ChipClassificationEvaluatorConfig()



from typing import TYPE_CHECKING, Iterable, Optional
from abc import (abstractmethod)
import logging

from rastervision.core.evaluation import Evaluator
from rastervision.core.data import Labels

log = logging.getLogger(__name__)

if TYPE_CHECKING:
    from rastervision.core.data import Scene, ClassConfig
    from rastervision.core.evaluation import ClassificationEvaluation


class ClassificationEvaluator(Evaluator):
    """Evaluates predictions for a set of scenes."""

    def __init__(self,
                 class_config: 'ClassConfig',
                 output_uri: Optional[str] = None):
        self.class_config = class_config
        self.output_uri = output_uri

    @abstractmethod
    def create_evaluation(self) -> 'ClassificationEvaluation':
        pass

    def process(self, scenes: Iterable['Scene'],
                tmp_dir: Optional[str] = None) -> None:
        if self.output_uri is not None:
            evaluation_global = self.create_evaluation()
            for scene in scenes:
                log.info(f'Computing evaluation for scene {scene.id}...')
                evaluation = self.evaluate_scene(scene)
                evaluation_global.merge(evaluation, scene_id=scene.id)
            evaluation_global.save(self.output_uri)

    def evaluate_scene(self, scene: 'Scene') -> 'ClassificationEvaluation':
        ground_truth = scene.label_source.get_labels()
        predictions = scene.label_store.get_labels()

        if scene.aoi_polygons:
            ground_truth = ground_truth.filter_by_aoi(scene.aoi_polygons)
            predictions = predictions.filter_by_aoi(scene.aoi_polygons)

        evaluation = self.evaluate_predictions(ground_truth, predictions)
        return evaluation

    def evaluate_predictions(
            self, ground_truth: 'Labels',
            predictions: 'Labels') -> 'ClassificationEvaluation':
        evaluation = self.create_evaluation()
        evaluation.compute(ground_truth, predictions)
        return evaluation



from rastervision.pipeline.config import Config, register_config


@register_config('label_store')
class LabelStoreConfig(Config):
    """Configure a :class:`.LabelStore`."""

    def build(self, class_config, crs_transformer, extent, tmp_dir):
        pass

    def update(self, pipeline=None, scene=None):
        pass



from typing import TYPE_CHECKING, Optional, Sequence, Union

from rastervision.core.data.utils.geojson import geoms_to_geojson

if TYPE_CHECKING:
    from rastervision.core.box import Box
    from rastervision.core.data import ClassConfig, CRSTransformer


def boxes_to_geojson(
        boxes: Sequence['Box'],
        class_ids: Sequence[int],
        crs_transformer: 'CRSTransformer',
        class_config: 'ClassConfig',
        scores: Optional[Sequence[Union[float, Sequence[float]]]] = None
) -> dict:
    """Convert boxes and associated data into a GeoJSON dict.

    Args:
        boxes (Sequence[Box]): List of Box in pixel row/col format.
        class_ids (Sequence[int]): List of int (one for each box)
        crs_transformer (CRSTransformer): CRSTransformer used to convert pixel
            coords to map coords in the GeoJSON.
        class_config (ClassConfig): ClassConfig
        scores (Optional[Sequence[Union[float, Sequence[float]]]], optional):
            Optional list of score or scores. If floats (one for each box),
            property name will be "score". If lists of floats, property name
            will be "scores". Defaults to None.

    Returns:
        dict: Serialized GeoJSON.
    """
    if len(boxes) != len(class_ids):
        raise ValueError(f'len(boxes) ({len(boxes)}) != '
                         f'len(class_ids) ({len(class_ids)})')
    if scores is not None and len(boxes) != len(scores):
        raise ValueError(f'len(boxes) ({len(boxes)}) != '
                         f'len(scores) ({len(scores)})')

    geoms = [crs_transformer.pixel_to_map(box.to_shapely()) for box in boxes]
    properties = [
        dict(class_id=id, class_name=class_config.get_name(id))
        for id in class_ids
    ]
    if scores is not None:
        for prop, box_score in zip(properties, scores):
            key = 'score' if isinstance(box_score, float) else 'scores'
            prop[key] = box_score

    geojson = geoms_to_geojson(geoms, properties)
    return geojson



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING
from contextlib import AbstractContextManager

if TYPE_CHECKING:
    from rastervision.core.data_sample import DataSample
    from rastervision.core.data import Labels, Scene


class SampleWriter(AbstractContextManager):
    """Writes DataSamples in a streaming fashion.

    This is a context manager used for creating training and validation chips, and
    should be subclassed for each Backend.
    """

    @abstractmethod
    def write_sample(self, sample: 'DataSample'):
        """Writes a single sample."""
        pass


class Backend(ABC):
    """Abstraction around core ML functionality used by an RVPipeline.

    This should be subclassed to enable use of a third party ML library with an
    RVPipeline. There is a one-to-many relationship from RVPipeline to Backend.
    """

    @abstractmethod
    def get_sample_writer(self):
        """Returns a SampleWriter for this Backend."""
        pass

    @abstractmethod
    def train(self):
        """Train a model.

        This should download chips created by the SampleWriter, train the model, and
        then saving it to disk.
        """
        pass

    @abstractmethod
    def load_model(self):
        """Load the model in preparation for one or more prediction calls."""
        pass

    @abstractmethod
    def predict_scene(self, scene: 'Scene', chip_sz: int,
                      stride: int) -> 'Labels':
        """Return predictions for an entire scene using the model.

        Args:
            scene (Scene): Scene to run inference on.

        Return:
            Labels object containing predictions
        """
        pass



from typing import TYPE_CHECKING, Iterable, Optional, Tuple
from os.path import join

from rastervision.pipeline.config import register_config, Config, Field

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig
    from rastervision.core.evaluation import Evaluator
    from rastervision.core.rv_pipeline import RVPipelineConfig


@register_config('evaluator')
class EvaluatorConfig(Config):
    """Configure an :class:`.Evaluator`."""

    output_uri: Optional[str] = Field(
        None,
        description='URI of directory where evaluator output will be saved. '
        'Evaluations for each scene-group will be save in a JSON file at '
        '<output_uri>/<scene-group-name>/eval.json. If None, and this Config '
        'is part of an RVPipeline, this field will be auto-generated.')

    def build(self,
              class_config: 'ClassConfig',
              scene_group: Optional[Tuple[str, Iterable[str]]] = None
              ) -> 'Evaluator':
        pass

    def get_output_uri(self, scene_group_name: Optional[str] = None) -> str:
        if scene_group_name is None:
            return join(self.output_uri, 'eval.json')
        return join(self.output_uri, scene_group_name, 'eval.json')

    def update(self, pipeline: Optional['RVPipelineConfig'] = None) -> None:
        if pipeline is not None and self.output_uri is None:
            self.output_uri = pipeline.eval_uri



from typing import Dict, List, Set

from rastervision.pipeline.config import (Config, register_config, ConfigError,
                                          Field)
from rastervision.pipeline.utils import split_into_groups
from rastervision.core.data.scene_config import SceneConfig
from rastervision.core.data.class_config import ClassConfig


def dataset_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version < 1:
        try:
            # removed in version 1
            del cfg_dict['img_channels']
        except KeyError:
            pass
    return cfg_dict


@register_config('dataset', upgrader=dataset_config_upgrader)
class DatasetConfig(Config):
    """Configure train, validation, and test splits for a dataset."""
    class_config: ClassConfig
    train_scenes: List[SceneConfig]
    validation_scenes: List[SceneConfig]
    test_scenes: List[SceneConfig] = []
    scene_groups: Dict[str, Set[str]] = Field(
        {},
        description='Groupings of scenes. Should be a dict of the form: '
        '{<group-name>: Set(scene_id_1, scene_id_2, ...)}. Three groups are '
        'added by default: "train_scenes", "validation_scenes", and '
        '"test_scenes"')

    def update(self, pipeline=None):
        super().update()

        self.class_config.update(pipeline=pipeline)
        for s in self.train_scenes:
            s.update(pipeline=pipeline)
        for s in self.validation_scenes:
            s.update(pipeline=pipeline)
        if self.test_scenes is not None:
            for s in self.test_scenes:
                s.update(pipeline=pipeline)

        # add default scene groups
        self.scene_groups['train_scenes'] = {s.id for s in self.train_scenes}
        self.scene_groups['test_scenes'] = {s.id for s in self.test_scenes}
        self.scene_groups['validation_scenes'] = {
            s.id
            for s in self.validation_scenes
        }

    def validate_config(self):
        ids = [s.id for s in self.train_scenes]
        if len(set(ids)) != len(ids):
            raise ConfigError('All training scene ids must be unique.')

        ids = [s.id for s in self.validation_scenes + self.test_scenes]
        if len(set(ids)) != len(ids):
            raise ConfigError(
                'All validation and test scene ids must be unique.')

        all_ids = {s.id for s in self.all_scenes}
        for group_name, group_ids in self.scene_groups.items():
            unknown_ids = group_ids.difference(all_ids)
            if len(unknown_ids) > 0:
                raise ConfigError(
                    f'IDs {unknown_ids} in scene group '
                    f'"{group_name}" do not match any scene in the dataset.')

    def get_split_config(self, split_ind, num_splits):
        new_cfg = self.copy()

        groups = split_into_groups(self.train_scenes, num_splits)
        new_cfg.train_scenes = groups[
            split_ind] if split_ind < len(groups) else []

        groups = split_into_groups(self.validation_scenes, num_splits)
        new_cfg.validation_scenes = groups[
            split_ind] if split_ind < len(groups) else []

        if self.test_scenes:
            groups = split_into_groups(self.test_scenes, num_splits)
            new_cfg.test_scenes = groups[
                split_ind] if split_ind < len(groups) else []

        return new_cfg

    @property
    def all_scenes(self) -> List[SceneConfig]:
        return self.train_scenes + self.validation_scenes + self.test_scenes



from typing import TYPE_CHECKING, Optional
from os.path import join

from rastervision.pipeline.config import register_config, Field
from rastervision.core.data.raster_transformer import (RasterTransformerConfig,
                                                       StatsTransformer)
from rastervision.core.raster_stats import RasterStats

if TYPE_CHECKING:
    from rastervision.core.rv_pipeline import RVPipelineConfig
    from rastervision.core.data import SceneConfig


def stats_transformer_config_upgrader(cfg_dict: dict, version: int) -> dict:
    if version == 2:
        # field added in version 3
        # since `scene_group` cannot be None, set it to a special value so that
        # `update_root()`, which is called by the predictor, knows to set
        # `stats_uri` to the old location of `stats.json`.
        cfg_dict['scene_group'] = '__N/A__'
    return cfg_dict


@register_config(
    'stats_transformer', upgrader=stats_transformer_config_upgrader)
class StatsTransformerConfig(RasterTransformerConfig):
    """Configure a :class:`.StatsTransformer`."""

    stats_uri: Optional[str] = Field(
        None,
        description='The URI of the output of the StatsAnalyzer. '
        'If None, and this Config is inside an RVPipeline, '
        'this field will be auto-generated.')
    scene_group: str = Field(
        'train_scenes',
        description='Name of the group of scenes whose stats to use. Defaults'
        'to "train_scenes".')

    def update(self,
               pipeline: Optional['RVPipelineConfig'] = None,
               scene: Optional['SceneConfig'] = None) -> None:
        if pipeline is not None and self.stats_uri is None:
            self.stats_uri = join(pipeline.analyze_uri, 'stats',
                                  self.scene_group, 'stats.json')

    def build(self):
        stats = RasterStats.load(self.stats_uri)
        return StatsTransformer(means=stats.means, stds=stats.stds)

    def update_root(self, root_dir: str) -> None:
        if self.scene_group == '__N/A__':
            # backward compatibility: use old location of stats.json
            self.stats_uri = join(root_dir, 'stats.json')
        else:
            self.stats_uri = join(root_dir, 'analyze', 'stats',
                                  self.scene_group, 'stats.json')



from typing import List
import logging

from rastervision.core.rv_pipeline.rv_pipeline import RVPipeline
from rastervision.core.rv_pipeline.utils import nodata_below_threshold
from rastervision.core.box import Box
from rastervision.core.data import Scene

log = logging.getLogger(__name__)


def get_train_windows(scene: Scene,
                      chip_size: int,
                      chip_nodata_threshold: float = 1.) -> List[Box]:
    train_windows = []
    extent = scene.raster_source.extent
    stride = chip_size
    windows = extent.get_windows(chip_size, stride)

    total_windows = len(windows)
    if scene.aoi_polygons:
        windows = Box.filter_by_aoi(windows, scene.aoi_polygons)
        log.info(f'AOI filtering: {len(windows)}/{total_windows} '
                 'chips accepted')
    for window in windows:
        chip = scene.raster_source.get_chip(window)
        if nodata_below_threshold(chip, chip_nodata_threshold, nodata_val=0):
            train_windows.append(window)
    log.info('NODATA filtering: '
             f'{len(train_windows)}/{len(windows)} chips accepted')
    return train_windows


class ChipClassification(RVPipeline):
    def get_train_windows(self, scene: Scene) -> List[Box]:
        return get_train_windows(
            scene,
            self.config.train_chip_sz,
            chip_nodata_threshold=self.config.chip_nodata_threshold)

    def get_train_labels(self, window: Box, scene: Scene):
        return scene.label_source.get_labels(window=window)



from typing import TYPE_CHECKING, Any, Iterable, List, Optional

import numpy as np
import geopandas as gpd

from rastervision.core.data.label import ChipClassificationLabels
from rastervision.core.data.label_source.label_source import LabelSource
from rastervision.core.box import Box

if TYPE_CHECKING:
    from rastervision.core.data import (ChipClassificationLabelSourceConfig,
                                        VectorSource)


def infer_cells(cells: List[Box], labels_df: gpd.GeoDataFrame,
                ioa_thresh: float, use_intersection_over_cell: bool,
                pick_min_class_id: bool,
                background_class_id: int) -> ChipClassificationLabels:
    """Infer ChipClassificationLabels grid from GeoJSON containing polygons.

    Given GeoJSON with polygons associated with class_ids, infer a grid of
    cells and class_ids that best captures the contents of each cell.

    For each cell, the problem is to infer the class_id that best captures the
    content of the cell. This is non-trivial since there can be multiple
    polygons of differing classes overlapping with the cell. Any polygons that
    sufficiently overlaps with the cell are in the running for setting the
    class_id. If there are none in the running, the cell is either
    considered null or background.

    Args:
        ioa_thresh: (float) the minimum IOA of a polygon and cell for that
            polygon to be a candidate for setting the class_id
        use_intersection_over_cell: (bool) If true, then use the area of the
            cell as the denominator in the IOA. Otherwise, use the area of the
            polygon.
        background_class_id: (None or int) If not None, class_id to use as the
            background class; ie. the one that is used when a window contains
            no boxes.
        pick_min_class_id: If true, the class_id for a cell is the minimum
            class_id of the boxes in that cell. Otherwise, pick the class_id of
            the box covering the greatest area.

    Returns:
        ChipClassificationLabels
    """
    cells_df = gpd.GeoDataFrame(
        data={'cell_id': range(len(cells))},
        geometry=[c.to_shapely() for c in cells])

    # duplicate geometry columns so that they are retained after the join
    cells_df.loc[:, 'geometry_cell'] = cells_df.geometry
    labels_df.loc[:, 'geometry_label'] = labels_df.geometry

    # Left-join cells to label polygons based on intersection. The result is a
    # table with each cell matched to all polygons that intersect it; i.e.,
    # there will be a row for each unique (cell, polygon) combination. Cells
    # that didn't match any labels will have missing values as their class_ids.
    df: gpd.GeoDataFrame = cells_df.sjoin(
        labels_df, how='left', predicate='intersects')
    df.loc[:, 'geometry_intersection'] = df['geometry_cell'].intersection(
        df['geometry_label'])

    if use_intersection_over_cell:
        ioa = (df['geometry_intersection'].area / df['geometry_cell'].area)
    else:
        # intersection over label-polygon
        ioa = (df['geometry_intersection'].area / df['geometry_label'].area)
    df.loc[:, 'ioa'] = ioa.fillna(-1)

    # labels with IOA below threshold cannot contribute their class_id
    df.loc[df['ioa'] < ioa_thresh, 'class_id'] = None

    # Assign background_class_id to cells /wo a class_id. This includes both
    # unmatched cells and ones whose ioa fell below the ioa_thresh.
    df.loc[df['class_id'].isna(), 'class_id'] = background_class_id

    # break ties (i.e. one cell matched to multiple label polygons)
    if pick_min_class_id:
        df = df.sort_values('class_id').drop_duplicates(
            ['cell_id'], keep='first')
    else:
        # largest IOA
        df = df.sort_values('ioa').drop_duplicates(['cell_id'], keep='last')

    boxes = [Box.from_shapely(c).to_int() for c in df['geometry_cell']]
    class_ids = df['class_id'].astype(int)
    cells_to_class_id = {
        cell: (class_id, None)
        for cell, class_id in zip(boxes, class_ids)
    }
    labels = ChipClassificationLabels(cells_to_class_id)
    return labels


def read_labels(labels_df: gpd.GeoDataFrame,
                extent: Optional[Box] = None) -> ChipClassificationLabels:
    """Convert GeoDataFrame to ChipClassificationLabels.

    If the GeoDataFrame already contains a grid of cells, then
    ChipClassificationLabels can be constructed in a straightforward manner
    without having to infer the class of cells.

    If extent is given, only labels that intersect with it are returned.

    Args:
        geojson: dict in normalized GeoJSON format (see VectorSource)
        extent: Box in pixel coords

    Returns:
       ChipClassificationLabels
    """
    if extent is not None:
        extent_polygon = extent.to_shapely()
        labels_df = labels_df[labels_df.intersects(extent_polygon)]
        boxes = np.array([
            Box.from_shapely(c).to_int().shift_origin(extent)
            for c in labels_df.geometry
        ])
    else:
        boxes = np.array(
            [Box.from_shapely(c).to_int() for c in labels_df.geometry])
    class_ids = labels_df['class_id'].astype(int)
    cells_to_class_id = {
        cell: (class_id, None)
        for cell, class_id in zip(boxes, class_ids)
    }
    labels = ChipClassificationLabels(cells_to_class_id)
    return labels


class ChipClassificationLabelSource(LabelSource):
    """A source of chip classification labels.

    Ideally the vector_source contains a square for each cell in the grid. But
    in reality, it can be difficult to label imagery in such an exhaustive way.
    So, this can also handle sources with non-overlapping polygons that
    do not necessarily cover the entire extent. It infers the grid of cells
    and associated class_ids using the extent and options if infer_cells is
    set to True.
    """

    def __init__(self,
                 label_source_config: 'ChipClassificationLabelSourceConfig',
                 vector_source: 'VectorSource',
                 extent: Box = None,
                 lazy: bool = False):
        """Constructs a LabelSource for chip classification.

        Args:
            label_source_config (ChipClassificationLabelSourceConfig): Config
                for class inference.
            vector_source (VectorSource): Source of vector labels.
            extent (Box): Box used to filter the labels by extent or
                compute grid.
            lazy (bool, optional): If True, labels are not populated during
                initialization. Defaults to False.
        """
        self.cfg = label_source_config
        self._extent = extent
        self.labels_df = vector_source.get_dataframe()
        self.validate_labels(self.labels_df)

        self.labels = ChipClassificationLabels.make_empty()
        if not lazy:
            self.populate_labels()

    def populate_labels(self, cells: Optional[Iterable[Box]] = None) -> None:
        """Populate self.labels by either reading or inferring.

        If cfg.infer_cells is True or specific cells are given, the labels are
        inferred. Otherwise, they are read from the geojson.
        """
        if self.cfg.infer_cells or cells is not None:
            self.labels = self.infer_cells(cells=cells)
        else:
            self.labels = read_labels(self.labels_df, extent=self.extent)

    def infer_cells(self, cells: Optional[Iterable[Box]] = None
                    ) -> ChipClassificationLabels:
        """Infer labels for a list of cells. Only cells whose labels are not
        already known are inferred.

        Args:
            cells (Optional[Iterable[Box]], optional): Cells whose labels are
                to be inferred. Defaults to None.

        Returns:
            ChipClassificationLabels: labels
        """
        cfg = self.cfg
        if cells is None:
            if cfg.cell_sz is None:
                raise ValueError('cell_sz is not set.')
            cells = self.extent.get_windows(cfg.cell_sz, cfg.cell_sz)
        else:
            cells = [cell.shift_origin(self.extent) for cell in cells]

        known_cells = [c for c in cells if c in self.labels]
        unknown_cells = [c for c in cells if c not in self.labels]

        labels = infer_cells(
            cells=unknown_cells,
            labels_df=self.labels_df,
            ioa_thresh=cfg.ioa_thresh,
            use_intersection_over_cell=cfg.use_intersection_over_cell,
            pick_min_class_id=cfg.pick_min_class_id,
            background_class_id=cfg.background_class_id)

        for cell in known_cells:
            class_id = self.labels.get_cell_class_id(cell)
            labels.set_cell(cell, class_id)

        return labels

    def get_labels(self,
                   window: Optional[Box] = None) -> ChipClassificationLabels:
        if window is None:
            return self.labels
        window = window.shift_origin(self.extent)
        return self.labels.get_singleton_labels(window)

    def __getitem__(self, key: Any) -> int:
        """Return label for a window, inferring it if it is not already known.
        """
        if isinstance(key, Box):
            window = key
            if window not in self.labels:
                self.labels += self.infer_cells(cells=[window])
            return self.labels[window].class_id
        else:
            return super().__getitem__(key)

    def validate_labels(self, df: gpd.GeoDataFrame) -> None:
        geom_types = set(df.geom_type)
        if 'Point' in geom_types or 'LineString' in geom_types:
            raise ValueError(
                'LineStrings and Points are not supported '
                'in ChipClassificationLabelSource. Use BufferTransformer '
                'to buffer them into Polygons.')

        if 'class_id' not in df.columns:
            raise ValueError('All label polygons must have a class_id.')

    @property
    def extent(self) -> Box:
        return self._extent



from rastervision.core.data.crs_transformer import CRSTransformer


class IdentityCRSTransformer(CRSTransformer):
    """Transformer for when map coordinates are already in pixel coordinates.

    This is useful for non-georeferenced imagery.
    """

    def _map_to_pixel(self, map_point):
        """Identity function.

        Args:
            map_point: (x, y) tuple in pixel coordinates

        Returns:
            (x, y) tuple in pixel coordinates
        """
        return map_point

    def _pixel_to_map(self, pixel_point):
        """Identity function.

        Args:
            pixel_point: (x, y) tuple in pixel coordinates

        Returns:
            (x, y) tuple in pixel coordinates
        """
        return pixel_point



from typing import TYPE_CHECKING, List, Optional
from os.path import join
import logging

from rastervision.pipeline import rv_config_ as rv_config
from rastervision.pipeline.config import (build_config, upgrade_config)
from rastervision.pipeline.file_system.utils import (download_if_needed,
                                                     file_to_json, unzip)
from rastervision.core.data.raster_source import ChannelOrderError
from rastervision.core.data import (SemanticSegmentationLabelStoreConfig,
                                    PolygonVectorOutputConfig,
                                    StatsTransformerConfig)
from rastervision.core.analyzer import StatsAnalyzerConfig

if TYPE_CHECKING:
    from rastervision.core.rv_pipeline import RVPipelineConfig
    from rastervision.core.data import SceneConfig

log = logging.getLogger(__name__)


class Predictor():
    """Class for making predictions based off of a model bundle."""

    def __init__(self,
                 model_bundle_uri: str,
                 tmp_dir: str,
                 update_stats: bool = False,
                 channel_order: Optional[List[int]] = None,
                 scene_group: Optional[str] = None):
        """Creates a new Predictor.

        Args:
            model_bundle_uri: URI of the model bundle to use. Can be any
                type of URI that Raster Vision can read.
            tmp_dir: Temporary directory in which to store files that are used
                by the Predictor. This directory is not cleaned up by this
                class.
            channel_order: Option for a new channel order to use for the
                imagery being predicted against. If not present, the
                channel_order from the original configuration in the predict
                package will be used.
        """
        self.tmp_dir = tmp_dir
        self.update_stats = update_stats
        self.model_loaded = False

        bundle_path = download_if_needed(model_bundle_uri)
        bundle_dir = join(tmp_dir, 'bundle')
        unzip(bundle_path, bundle_dir)

        config_path = join(bundle_dir, 'pipeline-config.json')
        config_dict = file_to_json(config_path)
        rv_config.set_everett_config(
            config_overrides=config_dict.get('rv_config'))
        config_dict = upgrade_config(config_dict)
        self.config: 'RVPipelineConfig' = build_config(config_dict)
        self.scene: 'SceneConfig' = self.config.dataset.validation_scenes[0]

        if not hasattr(self.scene.raster_source, 'uris'):
            raise Exception(
                'raster_source in model bundle must have uris as field')

        if not hasattr(self.scene.label_store, 'uri'):
            raise Exception(
                'label_store in model bundle must have uri as field')

        for t in self.scene.raster_source.transformers:
            if isinstance(t, StatsTransformerConfig):
                if scene_group is not None:
                    t.scene_group = scene_group
                else:
                    log.warn(f'Using stats for scene group "{t.scene_group}". '
                             'To use a different scene group, specify '
                             '--scene-group <scene-group-name>.')
            t.update_root(bundle_dir)

        if self.update_stats:
            stats_analyzer = StatsAnalyzerConfig(
                output_uri=join(bundle_dir, 'stats.json'))
            self.config.analyzers = [stats_analyzer]

        self.scene.label_source = None
        self.scene.aoi_uris = None

        self.config.dataset.train_scenes = [self.scene]
        self.config.dataset.validation_scenes = [self.scene]
        self.config.dataset.test_scenes = []
        self.config.train_uri = bundle_dir

        if channel_order is not None:
            self.scene.raster_source.channel_order = channel_order

        self.pipeline = None

    def predict(self, image_uris: List[str], label_uri: str) -> None:
        """Generate predictions for the given image.

        Args:
            image_uris: URIs of the images to make predictions against.
                This can be any type of URI readable by Raster Vision
                FileSystems.
            label_uri: URI to save labels off into
        """
        if self.pipeline is None:
            self.scene.raster_source.uris = image_uris
            self.pipeline = self.config.build(self.tmp_dir)
            if not hasattr(self.pipeline, 'predict'):
                raise Exception(
                    'pipeline in model bundle must have predict method')

        self.scene.raster_source.uris = image_uris
        self.scene.label_store.uri = label_uri

        if isinstance(self.scene.label_store,
                      SemanticSegmentationLabelStoreConfig):
            # create vector outputs for each class without specifying URIs
            self.scene.label_store.vector_output = [
                PolygonVectorOutputConfig(class_id=i)
                for i, _ in enumerate(self.config.dataset.class_config.names)
            ]
            # set URIs
            self.scene.label_store.uri = label_uri
            for vo in self.scene.label_store.vector_output:
                vo.update(self.config, self.scene, uri_prefix=label_uri)

        try:
            if self.update_stats:
                self.pipeline.analyze()
            self.pipeline.predict()
        except ChannelOrderError:
            raise ValueError(
                'The predict package is using a channel_order '
                'with channels unavailable in the imagery.\nTo set a new '
                'channel_order that only uses channels available in the '
                'imagery, use the --channel-order option.')



# flake8: noqa
from rastervision.core.data.raster_source.raster_source import *
from rastervision.core.data.raster_source.raster_source_config import *
from rastervision.core.data.raster_source.rasterio_source import *
from rastervision.core.data.raster_source.rasterio_source_config import *
from rastervision.core.data.raster_source.rasterized_source import *
from rastervision.core.data.raster_source.rasterized_source_config import *
from rastervision.core.data.raster_source.multi_raster_source import *
from rastervision.core.data.raster_source.multi_raster_source_config import *

__all__ = [
    RasterSource.__name__,
    RasterSourceConfig.__name__,
    RasterioSource.__name__,
    RasterioSourceConfig.__name__,
    RasterizedSource.__name__,
    RasterizedSourceConfig.__name__,
    RasterizerConfig.__name__,
    MultiRasterSource.__name__,
    MultiRasterSourceConfig.__name__,
]



from pydantic import BaseModel
from numpy import ndarray

from rastervision.core.box import Box
from rastervision.core.data import Labels


class DataSample(BaseModel):
    """A chip and labels along with metadata."""
    chip: ndarray
    window: Box
    labels: Labels
    scene_id: str = 'default'
    is_train: bool = True

    class Config:
        arbitrary_types_allowed = True



from typing import TYPE_CHECKING, Iterable, Optional, Tuple

from rastervision.pipeline.config import register_config
from rastervision.core.evaluation.classification_evaluator_config import (
    ClassificationEvaluatorConfig)
from rastervision.core.evaluation.object_detection_evaluator import (
    ObjectDetectionEvaluator)

if TYPE_CHECKING:
    from rastervision.core.data import ClassConfig


@register_config('object_detection_evaluator')
class ObjectDetectionEvaluatorConfig(ClassificationEvaluatorConfig):
    """Configure an :class:`.ObjectDetectionEvaluator`."""

    def build(self,
              class_config: 'ClassConfig',
              scene_group: Optional[Tuple[str, Iterable[str]]] = None
              ) -> ObjectDetectionEvaluator:
        if scene_group is None:
            output_uri = self.get_output_uri()
        else:
            group_name, _ = scene_group
            output_uri = self.get_output_uri(group_name)

        return ObjectDetectionEvaluator(class_config, output_uri)



from abc import (ABC, abstractmethod)
from typing import TYPE_CHECKING, Optional

if TYPE_CHECKING:
    from rastervision.core.data import CRSTransformer


class VectorTransformer(ABC):
    """Transforms vector data."""

    def __call__(self,
                 geojson: dict,
                 crs_transformer: Optional['CRSTransformer'] = None,
                 **kwargs) -> dict:
        return self.transform(
            geojson, crs_transformer=crs_transformer, **kwargs)

    @abstractmethod
    def transform(self,
                  geojson: dict,
                  crs_transformer: Optional['CRSTransformer'] = None) -> dict:
        """Transform a GeoJSON mapping of vector data.

        Args:
            geojson (dict): A GeoJSON-like mapping of a FeatureCollection.

        Returns:
            dict: Transformed GeoJSON.
        """
        pass



# flake8: noqa

TRAIN = 'train'
VALIDATION = 'validation'

from rastervision.core.rv_pipeline.rv_pipeline import *
from rastervision.core.rv_pipeline.rv_pipeline_config import *
from rastervision.core.rv_pipeline.chip_classification import *
from rastervision.core.rv_pipeline.chip_classification_config import *
from rastervision.core.rv_pipeline.semantic_segmentation import *
from rastervision.core.rv_pipeline.semantic_segmentation_config import *
from rastervision.core.rv_pipeline.object_detection import *
from rastervision.core.rv_pipeline.object_detection_config import *
from rastervision.core.rv_pipeline.utils import *

__all__ = [
    RVPipeline.__name__,
    RVPipelineConfig.__name__,
    ChipClassification.__name__,
    ChipClassificationConfig.__name__,
    SemanticSegmentation.__name__,
    SemanticSegmentationConfig.__name__,
    SemanticSegmentationChipOptions.__name__,
    SemanticSegmentationPredictOptions.__name__,
    SemanticSegmentationWindowMethod.__name__,
    ObjectDetection.__name__,
    ObjectDetectionConfig.__name__,
    ObjectDetectionChipOptions.__name__,
    ObjectDetectionChipOptions.__name__,
    ObjectDetectionPredictOptions.__name__,
]



from typing import TYPE_CHECKING, Iterator, Optional, Sequence
import json

import numpy as np
from tqdm.auto import tqdm

from rastervision.pipeline.file_system import str_to_file, file_to_str

if TYPE_CHECKING:
    from rastervision.core.box import Box
    from rastervision.core.data import RasterSource

chip_sz = 300


def parallel_variance(mean_a, count_a, var_a, mean_b, count_b, var_b):
    """Compute the variance based on stats from two partitions of the data.

    See "Parallel Algorithm" in
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance

    Args:
        mean_a: the mean of partition a
        count_a: the number of elements in partition a
        var_a: the variance of partition a
        mean_b: the mean of partition b
        count_b: the number of elements in partition b
        var_b: the variance of partition b

    Return:
        the variance of the two partitions if they were combined
    """
    delta = mean_b - mean_a
    m_a = var_a * (count_a - 1)
    m_b = var_b * (count_b - 1)
    M2 = m_a + m_b + delta**2 * count_a * count_b / (count_a + count_b)
    var = M2 / (count_a + count_b - 1)
    return var


def parallel_mean(mean_a, count_a, mean_b, count_b):
    """Compute the mean based on stats from two partitions of the data.

    See "Parallel Algorithm" in
    https://en.wikipedia.org/wiki/Algorithms_for_calculating_variance

    Args:
        mean_a: the mean of partition a
        count_a: the number of elements in partition a
        mean_b: the mean of partition b
        count_b: the number of elements in partition b

    Return:
        the mean of the two partitions if they were combined
    """
    mean = (count_a * mean_a + count_b * mean_b) / (count_a + count_b)
    return mean


class RasterStats():
    def __init__(self):
        self.means = None
        self.stds = None

    def compute(self,
                raster_sources: Sequence['RasterSource'],
                sample_prob: Optional[float] = None) -> None:
        """Compute the mean and stds over all the raster_sources.

        This ignores NODATA values.

        If sample_prob is set, then a subset of each scene is used to compute stats which
        speeds up the computation. Roughly speaking, if sample_prob=0.5, then half the
        pixels in the scene will be used. More precisely, the number of chips is equal to
        sample_prob * (width * height / 300^2), or 1, whichever is greater. Each chip is
        uniformly sampled from the scene with replacement. Otherwise, it uses a sliding
        window over the entire scene to compute stats.

        Args:
            raster_sources: list of RasterSource
            sample_prob: (float or None) between 0 and 1
        """
        stride = chip_sz
        nb_channels = raster_sources[0].num_channels_raw

        def get_chip(raster_source: 'RasterSource',
                     window: 'Box') -> Optional[np.ndarray]:
            """Return chip or None if all values are NODATA."""
            chip = raster_source.get_raw_chip(window).astype(float)
            # Convert shape from [h,w,c] to [c,h*w]
            chip = np.reshape(np.transpose(chip, [2, 0, 1]), (nb_channels, -1))

            # Ignore NODATA values.
            chip[chip == 0.0] = np.nan
            if np.any(~np.isnan(chip)):
                return chip
            return None

        def sliding_chip_stream() -> Iterator[np.ndarray]:
            """Get stream of chips using a sliding window of size 300."""
            for raster_source in raster_sources:
                windows = raster_source.extent.get_windows(chip_sz, stride)
                for window in windows:
                    chip = get_chip(raster_source, window)
                    if chip is not None:
                        yield chip

        def random_chip_stream() -> Iterator[np.ndarray]:
            """Get random stream of chips."""
            for raster_source in raster_sources:
                extent = raster_source.extent
                num_pixels = extent.area
                num_chips = round(sample_prob * (num_pixels / (chip_sz**2)))
                num_chips = max(1, num_chips)
                for _ in range(num_chips):
                    window = raster_source.extent.make_random_square(chip_sz)
                    chip = get_chip(raster_source, window)
                    if chip is not None:
                        yield chip

        # For each chip, compute the mean and var of that chip and then update the
        # running mean and var.
        count = 0
        mean = np.zeros((nb_channels, ))
        var = np.zeros((nb_channels, ))
        chip_stream = (sliding_chip_stream()
                       if sample_prob is None else random_chip_stream())

        with tqdm(chip_stream, desc='Analyzing chips') as bar:
            for chip in bar:
                chip_means = np.nanmean(chip, axis=1)
                chip_vars = np.nanvar(chip, axis=1)
                chip_count = np.sum(chip[0] != np.nan)

                var = parallel_variance(chip_means, chip_count, chip_vars,
                                        mean, count, var)
                mean = parallel_mean(chip_means, chip_count, mean, count)
                count += chip_count

        self.means = mean
        self.stds = np.sqrt(var)

    def save(self, stats_uri: str) -> None:
        # Ensure lists
        means = list(self.means)
        stds = list(self.stds)
        stats = {'means': means, 'stds': stds}
        str_to_file(json.dumps(stats), stats_uri)

    @staticmethod
    def load(stats_uri: str) -> None:
        stats_json = json.loads(file_to_str(stats_uri))
        stats = RasterStats()
        stats.means = stats_json['means']
        stats.stds = stats_json['stds']
        return stats



"""Defines ``ClassEvaluationItem``."""

from typing import Optional

import numpy as np

from rastervision.core.evaluation import EvaluationItem


class ClassEvaluationItem(EvaluationItem):
    """A wrapper around a binary (2x2) confusion matrix of the form

    .. line-block::
       [``TN`` ``FP``]
       [``FN`` ``TP``]

    where ``TN`` need not necessarily be available.

    Exposes evaluation metrics computed from the confusion matrix as
    properties.

    Attributes:
        class_id (int): Class ID.
        class_name (str): Class name.
        conf_mat (np.ndarray): Confusion matrix: ``[[TN, FP], [FN, TP]]``.
        extra_info (dict): Arbitrary extra key-value pairs that will be
            included in the dict returned by ``to_json()``.
    """

    def __init__(self,
                 class_id: int,
                 class_name: str,
                 tp: int,
                 fp: int,
                 fn: int,
                 tn: Optional[int] = None,
                 **kwargs):
        """Constructor.

        Args:
            class_id (int): Class ID.
            class_name (str): Class name.
            tp (int): True positive count.
            fp (int): False positive count.
            fn (int): False negative count.
            tn (Optional[int], optional): True negative count.
                Defaults to None.
            **kwargs: Additional data can be provided as keyword arguments.
                These will be included in the dict returned by ``to_json()``.
        """
        self.class_id = class_id
        self.class_name = class_name
        if tn is None:
            tn = -1
        self.conf_mat = np.array([[tn, fp], [fn, tp]])
        self.extra_info = kwargs

    @classmethod
    def from_multiclass_conf_mat(cls, conf_mat: np.ndarray, class_id: int,
                                 class_name: str,
                                 **kwargs) -> 'ClassEvaluationItem':
        """Construct from a multi-class confusion matrix and a target class ID.

        Args:
            conf_mat (np.ndarray): A multi-class confusion matrix.
            class_id (int): The ID of the target class.
            class_name (str): The name of the target class.

        Returns:
            ClassEvaluationItem: ClassEvaluationItem for target class.
        """
        tp = conf_mat[class_id, class_id]
        fp = conf_mat[:, class_id].sum() - tp
        fn = conf_mat[class_id, :].sum() - tp
        tn = conf_mat.sum() - tp - fp - fn
        item = cls(
            class_id=class_id,
            class_name=class_name,
            tp=tp,
            fp=fp,
            fn=fn,
            tn=tn,
            **kwargs)
        return item

    def merge(self, other: 'ClassEvaluationItem') -> None:
        """Merge with another ``ClassEvaluationItem``.

        This is accomplished by summing the confusion matrices.
        """
        if self.class_id != other.class_id:
            raise ValueError(
                'Cannot merge evaluation items for different classes.')
        self.conf_mat += other.conf_mat

    @property
    def gt_count(self) -> int:
        """Positive ground-truth count."""
        return self.conf_mat[1, :].sum()

    @property
    def pred_count(self) -> int:
        """Positive prediction count."""
        return self.conf_mat[:, 1].sum()

    @property
    def true_pos(self) -> int:
        """True positive count."""
        return self.conf_mat[1, 1]

    @property
    def true_neg(self) -> Optional[int]:
        """True negative count.

        Returns:
            Optional[int]: Count as int if available. Otherwise, None.
        """
        tn = self.conf_mat[0, 0]
        if tn < 0:
            return None
        return tn

    @property
    def false_pos(self) -> int:
        """False positive count."""
        return self.conf_mat[0, 1]

    @property
    def false_neg(self) -> int:
        """False negative count."""
        return self.conf_mat[1, 0]

    @property
    def recall(self) -> float:
        """``TP / (TP + FN)``"""
        tp = self.true_pos
        fn = self.false_neg
        return float(tp) / (tp + fn)

    @property
    def sensitivity(self) -> float:
        """Equivalent to ``recall``."""
        return self.recall

    @property
    def specificity(self) -> Optional[float]:
        """``TN / (TN + FP)``"""
        if self.true_neg is None:
            return None
        tn = self.true_neg
        fp = self.false_pos
        return float(tn) / (tn + fp)

    @property
    def precision(self) -> float:
        """``TP / (TP + FP)``"""
        tp = self.true_pos
        fp = self.false_pos
        return float(tp) / (tp + fp)

    @property
    def f1(self) -> float:
        """F1 score = ``2 * (precision * recall) / (precision + recall)``"""
        precision = self.precision
        recall = self.recall
        return 2 * (precision * recall) / (precision + recall)

    def to_json(self) -> dict:
        """Serialize to a dict."""
        out = {
            'class_id': self.class_id,
            'class_name': self.class_name,
            'gt_count': self.gt_count,
            'pred_count': self.pred_count,
            'count_error': abs(self.gt_count - self.pred_count),
            'relative_frequency': self.gt_count / self.conf_mat.sum(),
            'metrics': {
                'recall': self.recall,
                'precision': self.precision,
                'f1': self.f1,
                'sensitivity': self.sensitivity,
                'specificity': self.specificity,
            }
        }
        if self.true_neg is None:
            del out['relative_frequency']
            out['true_pos'] = self.true_pos
            out['false_pos'] = self.false_pos
            out['false_neg'] = self.false_neg
        else:
            cm = self.conf_mat
            cm_frac = cm / cm.sum()
            out['conf_mat'] = cm.tolist()
            out['conf_mat_frac'] = cm_frac.tolist()
            [[TN, FP], [FN, TP]] = cm
            out['conf_mat_dict'] = dict(TN=TN, FP=FP, FN=FN, TP=TP)
            [[TN, FP], [FN, TP]] = cm_frac
            out['conf_mat_frac_dict'] = dict(TN=TN, FP=FP, FN=FN, TP=TP)

        out.update(self.extra_info)
        return out



from typing import List, Optional
from urllib.parse import urlparse
import logging
from itertools import islice

import boto3
from pystac import StacIO, Catalog, Item
from pystac.stac_io import DefaultStacIO
from shapely.geometry import box

from rastervision.pipeline.file_system import file_to_json

log = logging.getLogger(__name__)


def setup_stac_io() -> None:
    def read_method(uri: str):
        parsed = urlparse(uri)
        if parsed.scheme == 's3':
            bucket = parsed.netloc
            key = parsed.path[1:]
            s3 = boto3.resource('s3')
            obj = s3.Object(bucket, key)
            return obj.get()['Body'].read().decode('utf-8')
        else:
            return DefaultStacIO.read_text_method(uri)

    def write_method(uri: str, txt: str):
        parsed = urlparse(uri)
        if parsed.scheme == 's3':
            bucket = parsed.netloc
            key = parsed.path[1:]
            s3 = boto3.resource('s3')
            s3.Object(bucket, key).put(Body=txt)
        else:
            DefaultStacIO.write_text_method(uri, txt)

    StacIO.read_text_method = read_method
    StacIO.write_text_method = write_method


def is_label_item(item: Item) -> bool:
    """Resolve each extension schema into a dict, then check if it has the
    title of "Label Extension".
    """
    for ext_schema_uri in item.stac_extensions:
        schema = file_to_json(ext_schema_uri)
        if schema['title'].lower() == 'label extension':
            return True
    return False


def get_linked_image_item(label_item: Item) -> Optional[Item]:
    """Find link in the item that has "rel" == "source" and return its
    "target" item. If no such link, return None. If multiple such links,
    raise an exception."""
    links = [link for link in label_item.links if link.rel.lower() == 'source']
    if len(links) == 0:
        return None
    elif len(links) > 1:
        raise NotImplementedError()
    image_item = links[0].resolve_stac_object().target
    return image_item


def parse_stac(stac_uri: str, item_limit: Optional[int] = None) -> List[dict]:
    """Parse a STAC catalog JSON file to extract label URIs, images URIs,
    and AOIs.

    Note: This has been tested to be compatible with STAC version 1.0.0 but
    not any other versions.

    Args:
        stac_uri (str): Path to the STAC catalog JSON file.

    Returns:
        List[dict]: A lsit of dicts with keys: "label_uri", "image_uris",
            "label_bbox", "image_bbox", "bboxes_intersect", and "aoi_geometry".
            Each dict corresponds to one label item and its associated image
            assets in the STAC catalog.
    """
    setup_stac_io()
    cat = Catalog.from_file(stac_uri)
    version: str = cat.to_dict()['stac_version']

    if not version.startswith('1.0'):
        log.warning(f'Parsing is not guaranteed to work correctly for '
                    f'STAC version != 1.0.*. Found version: {version}.')

    cat.make_all_asset_hrefs_absolute()

    label_items = list(
        islice(filter(is_label_item, cat.get_all_items()), item_limit))
    image_items = [get_linked_image_item(item) for item in label_items]

    if len(label_items) == 0:
        raise ValueError('Unable to find any label items in STAC catalog.')

    out = []
    for label_item, image_item in zip(label_items, image_items):
        label_uri: str = list(label_item.assets.values())[0].href
        label_bbox = box(*label_item.bbox)
        aoi_geometry: Optional[dict] = label_item.geometry

        if image_item is not None:
            image_assets = [
                asset for asset in image_item.get_assets().values()
                if 'image' in asset.media_type
            ]
            image_uris = [asset.href for asset in image_assets]
            image_bbox = box(*image_item.bbox)
            bboxes_intersect = label_bbox.intersects(image_bbox)
        else:
            image_uris = []
            image_bbox = None
            bboxes_intersect = False

        out.append({
            'label_uri': label_uri,
            'image_uris': image_uris,
            'label_bbox': label_bbox,
            'image_bbox': image_bbox,
            'bboxes_intersect': bboxes_intersect,
            'aoi_geometry': aoi_geometry
        })
    return out


def read_stac(uri: str, extract_dir: Optional[str] = None,
              **kwargs) -> List[dict]:
    """Parse the contents of a STAC catalog (downloading it first, if
    remote). If the uri is a zip file, unzip it, find catalog.json inside it
    and parse that.

    Args:
        uri (str): Either a URI to a STAC catalog JSON file or a URI to a zip
            file containing a STAC catalog JSON file.

    Raises:
        FileNotFoundError: If catalog.json is not found inside the zip file.
        Exception: If multiple catalog.json's are found inside the zip file.

    Returns:
        List[dict]: A lsit of dicts with keys: "label_uri", "image_uris",
            "label_bbox", "image_bbox", "bboxes_intersect", and "aoi_geometry".
            Each dict corresponds to one label item and its associated image
            assets in the STAC catalog.
    """
    from pathlib import Path
    from rastervision.pipeline.file_system.utils import (download_if_needed,
                                                         is_archive, extract)

    catalog_path = download_if_needed(uri)
    if catalog_path.lower().endswith('.json'):
        return parse_stac(catalog_path, **kwargs)

    if not is_archive(catalog_path):
        raise ValueError(f'Unsupported file format: ("{uri}"). '
                         'URIS must be a JSON file or compressed archive.')

    extract_dir = extract(catalog_path, extract_dir)
    catalog_paths = list(Path(extract_dir).glob('**/catalog.json'))
    if len(catalog_paths) == 0:
        raise FileNotFoundError(f'Unable to find "catalog.json" in {uri}.')
    elif len(catalog_paths) > 1:
        raise Exception(f'More than one "catalog.json" found in ' f'{uri}.')
    catalog_path = str(catalog_paths[0])
    return parse_stac(catalog_path, **kwargs)



from typing import List, Optional, Tuple, Union

from rastervision.pipeline.config import (Config, register_config, ConfigError,
                                          Field, validator)
from rastervision.core.data.utils import color_to_triple, normalize_color

DEFAULT_NULL_CLASS_NAME = 'null'
DEFAULT_NULL_CLASS_COLOR = 'black'


@register_config('class_config')
class ClassConfig(Config):
    """Configure class information for a machine learning task."""

    names: List[str] = Field(
        ...,
        description='Names of classes. The i-th class in this list will have '
        'class ID = i.')
    colors: Optional[List[Union[str, Tuple]]] = Field(
        None,
        description=
        ('Colors used to visualize classes. Can be color strings accepted by '
         'matplotlib or RGB tuples. If None, a random color will be auto-generated '
         'for each class.'))
    null_class: Optional[str] = Field(
        None,
        description='Optional name of class in `names` to use as the null '
        'class. This is used in semantic segmentation to represent the label '
        'for imagery pixels that are NODATA or that are missing a label. '
        f'If None and the class names include "{DEFAULT_NULL_CLASS_NAME}", '
        'it will automatically be used as the null class. If None, and this '
        'Config is part of a SemanticSegmentationConfig, a null class will be '
        'added automatically.')

    @validator('colors', always=True)
    def validate_colors(cls, v: Optional[List[Union[str, Tuple]]],
                        values: dict) -> Optional[List[Union[str, Tuple]]]:
        """Compare length w/ names. Also auto-generate if not specified."""
        class_names = values['names']
        class_colors = v
        if class_colors is None:
            class_colors = [color_to_triple() for _ in class_names]
        elif len(class_names) != len(class_colors):
            raise ConfigError(f'len(class_names) ({len(class_names)}) != '
                              f'len(class_colors) ({len(class_colors)})\n'
                              f'class_names: {class_names}\n'
                              f'class_colors: {class_colors}')
        return class_colors

    @validator('null_class', always=True)
    def validate_null_class(cls, v: Optional[str],
                            values: dict) -> Optional[str]:
        """Check if in names. If 'null' in names, use it as null class."""
        names = values['names']
        if v is None:
            if DEFAULT_NULL_CLASS_NAME in names:
                v = DEFAULT_NULL_CLASS_NAME
        else:
            if v not in names:
                raise ConfigError(
                    f'The null_class, "{v}", must be in list of class names.')

            # edge case
            default_null_class_in_names = (DEFAULT_NULL_CLASS_NAME in names)
            null_class_neq_default = (v != DEFAULT_NULL_CLASS_NAME)
            if default_null_class_in_names and null_class_neq_default:
                raise ConfigError(
                    f'"{DEFAULT_NULL_CLASS_NAME}" is in names but the '
                    f'specified null_class is something else ("{v}").')
        return v

    def get_class_id(self, name: str) -> int:
        return self.names.index(name)

    def get_name(self, id: int) -> str:
        return self.names[id]

    @property
    def null_class_id(self) -> int:
        if self.null_class is None:
            raise ValueError('null_class is not set')
        return self.get_class_id(self.null_class)

    def get_color_to_class_id(self) -> dict:
        return dict([(self.colors[i], i) for i in range(len(self.colors))])

    def ensure_null_class(self) -> None:
        """Add a null class if one isn't set. This method is idempotent."""
        if self.null_class is not None:
            return

        null_class_name = DEFAULT_NULL_CLASS_NAME
        null_class_color = DEFAULT_NULL_CLASS_COLOR

        # This might seeem redundant given the null class validator above, but
        # is actually important. Sometimes there can be multiple ClassConfig
        # instances that reference the same list objects for names and colors
        # (not clear why this happens). This means that
        # each ensure_null_class() call will add to names and colors in each
        # copy of ClassConfig but only set its own null_class, which makes this
        # method() non-idempotent.
        if null_class_name in self.names:
            self.null_class = null_class_name
            return

        # use random color if default color is already taken
        null_class_color_triple = color_to_triple(null_class_color)
        all_color_triples = [
            color_to_triple(c) if isinstance(c, str) else c
            for c in self.colors
        ]
        if null_class_color_triple in all_color_triples:
            null_class_color = color_to_triple()

        self.names.append(null_class_name)
        self.colors.append(null_class_color)
        self.null_class = null_class_name

    def __len__(self) -> int:
        return len(self.names)

    @property
    def color_triples(self) -> List[Tuple[float, float, float]]:
        """Class colors in a normalized form."""
        color_triples = [normalize_color(c) for c in self.colors]
        return color_triples
